/* Implement actions for CHILL.
   Copyright (C) 1992, 93, 1994 Free Software Foundation, Inc.
   Authors: Per Bothner, Bill Cox, Michael Tiemann, Michael North

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <stdio.h>
#include <limits.h>
#include "config.h"
#include "tree.h"
#include "rtl.h"
#include "expr.h"
#include "ch-tree.h"
#include "lex.h"
#include "flags.h"
#include "actions.h"
#include "obstack.h"
#include "assert.h"

#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free

/* reserved tag definitions */

#define TYPE_ID                 "id"
#define TAG_OBJECT              "chill_object"
#define TAG_CLASS               "chill_class"

extern int flag_short_enums;
extern int current_nesting_level;

extern rtx  assign_stack_temp         PROTO((enum machine_mode, int, int));
extern tree build_chill_compound_expr PROTO((tree));
extern tree build_chill_exception_decl PROTO((char *));
extern tree convert                   PROTO((tree, tree));
extern rtx  emit_line_note_force      PROTO((char *, int));
extern void error                     PROTO((char *, ...));
extern void error_with_decl           PROTO((tree, char *, ...));
extern rtx  gen_nop                   PROTO((void));
extern tree get_identifier            PROTO((char *));
extern void pedwarn                   PROTO((char *, ...));
extern tree size_int                  PROTO((unsigned int));
extern void sorry                     PROTO((char *, ...));
extern void warning                   PROTO((char *, ...));

extern tree   instance_type_node;
extern tree   sizetype;
extern struct obstack *expression_obstack, permanent_obstack;
extern struct obstack *current_obstack, *saveable_obstack;

/* This flag is checked throughout the non-CHILL-specific
   in the front end. */
tree chill_integer_type_node;
tree chill_unsigned_type_node;

/* Never used.  Referenced from c-typeck.c, which we use. */
int current_function_returns_value = 0;
int current_function_returns_null = 0;

/* data imported from toplev.c  */

extern char *dump_base_name;

/* set from command line parameter, to exit after 
   grant file written, generating no code. */
int grant_only_flag = 0;

char *
lang_identify ()
{
  return "chill";
}


void
init_chill ()
{
}


void
print_lang_statistics ()
{
}


void
lang_finish ()
{
#if 0
    extern int errorcount, sorrycount;

    /* this should be the last action in compiling a module.
       If there are other actions to be performed at lang_finish
       please insert before this */

    /* FIXME: in case of a syntax error, this leaves the grant file incomplete */
    /* for the moment we print a warning in case of errors and 
       continue granting */
    if ((errorcount || sorrycount) && grant_count)
      {
	warning ("%d errors, %d sorries, do granting", errorcount, sorrycount);
	errorcount = sorrycount = 0;
      }
#endif
}

void
chill_check_decl (decl)
     tree decl;
{
  tree type = TREE_TYPE (decl);
  static int alreadyWarned = 0;

  if (TREE_CODE (type) == RECORD_TYPE) /* && TREE_STATIC_TEMPLATE (type)) */
    {
      if (!alreadyWarned)
        {
          error ("GNU compiler does not support statically allocated objects");          
          alreadyWarned = 1;
        }
      error_with_decl (decl, "`%s' cannot be statically allocated");
    }
}

/* Comparison function for sorting identifiers in RAISES lists.
   Note that because IDENTIFIER_NODEs are unique, we can sort
   them by address, saving an indirection.  */
static int
id_cmp (p1, p2)
     tree *p1, *p2;
{
  return (int)TREE_VALUE (*p1) - (int)TREE_VALUE (*p2);
}

/* Build the FUNCTION_TYPE or METHOD_TYPE which may raise exceptions
   listed in RAISES.  */
tree
build_exception_variant (type, raises)
     tree type, raises;
{
  int i;
  tree v = TYPE_MAIN_VARIANT (type);
  tree t, t2;
  int constp    = TYPE_READONLY (type);
  int volatilep = TYPE_VOLATILE (type);

  if (!raises)
    return build_type_variant (v, constp, volatilep);

  if (TREE_CHAIN (raises))
    { /* Sort the list */
      tree *a = (tree *)alloca ((list_length (raises)+1) * sizeof (tree));
      for (i = 0, t = raises; t; t = TREE_CHAIN (t), i++)
	a[i] = t;
      /* NULL terminator for list.  */
      a[i] = NULL_TREE;
      qsort (a, i, sizeof (tree), id_cmp);
      while (i--)
	TREE_CHAIN (a[i]) = a[i+1];
      raises = a[0];
    }

  for (v = TYPE_NEXT_VARIANT (v); v; v = TYPE_NEXT_VARIANT (v))
    {
      if (TYPE_READONLY (v) != constp
	  || TYPE_VOLATILE (v) != volatilep)
	continue;

      t = raises;
      t2 = TYPE_RAISES_EXCEPTIONS (v);
      while (t && t2)
	{
	  if (TREE_TYPE (t) == TREE_TYPE (t2))
	    {
	      t = TREE_CHAIN (t);
	      t2 = TREE_CHAIN (t2);
	    }
	  else break;
	}
      if (t || t2)
	continue;
      /* List of exceptions raised matches previously found list.

         @@ Nice to free up storage used in consing up the
	 @@ list of exceptions raised.  */
      return v;
    }

  /* Need to build a new variant.  */
  if (TREE_PERMANENT (type))
    {
      push_obstacks_nochange ();
      end_temporary_allocation ();
      v = copy_node (type);
      pop_obstacks ();
    }
  else
    v = copy_node (type);

  TYPE_NEXT_VARIANT (v) = TYPE_NEXT_VARIANT (type);
  TYPE_NEXT_VARIANT (type) = v;
  if (raises && ! TREE_PERMANENT (raises))
    {
      push_obstacks_nochange ();
      end_temporary_allocation ();
      raises = copy_list (raises);
      pop_obstacks ();
    }
  TYPE_RAISES_EXCEPTIONS (v) = raises;
  return v;
}
#if 0

tree
build_rts_call (name, type, args)
     char *name;
     tree type, args;
{
  tree decl = lookup_name (get_identifier (name));
  tree converted_args = NULL_TREE;
  tree result, length = NULL_TREE;

  assert (decl != NULL_TREE);
  while (args)
    {
      tree arg = TREE_VALUE (args);
      if (TREE_CODE (TREE_TYPE (arg)) == SET_TYPE
	  || TREE_CODE (TREE_TYPE (arg)) == ARRAY_TYPE)
	{
	  length = size_in_bytes (TREE_TYPE (arg));
	  arg = build_chill_addr_expr (arg, (char *)0);
	}
      converted_args = tree_cons (NULL_TREE, arg, converted_args);
      args = TREE_CHAIN (args);
    }
  if (length != NULL_TREE)
    converted_args = tree_cons (NULL_TREE, length, converted_args);
  converted_args = nreverse (converted_args);
  result = build_chill_function_call (decl, converted_args);
  if (TREE_CODE (type) == SET_TYPE || TREE_CODE (type) == ARRAY_TYPE)
    result = build1 (INDIRECT_REF, type, result);
  else
    result = convert (type, result);
  return result;
}
#endif

/*
 * build a call to the following function:
 *   void   __cause_exception (char** ex, const char *file, 
 *                             const unsigned lineno);
 */
tree
build_cause_exception (exp_name)
     char *exp_name;
{
  /* We don't use build_rts_call() here, because the string (array of char)
     would be followed by its length in the parameter list built by
     build_rts_call, and the runtime routine doesn't want a length parameter.*/
  tree exp_decl = build_chill_exception_decl (exp_name);
  tree function = lookup_name (get_identifier ("__cause_exception"));
  tree result =
    build_chill_function_call (function,
      tree_cons (NULL_TREE, build_chill_addr_expr (exp_decl, (char *)0),
	tree_cons (NULL_TREE, 
          force_addr_of (get_chill_filename ()),
	    tree_cons (NULL_TREE, get_chill_linenumber (), NULL_TREE))));
  return result;
}

void
expand_cause_exception (exp_name)
     char *exp_name;
{
  expand_expr_stmt (build_cause_exception (exp_name));
}


/*
 * generate a conditional exception, if the given
 * condition is TRUE.  Should we check a command-line
 * flag to enable, disable exception generation?
 * If no condition specified, just CAUSE the !@#$%^ exception.
 */
void
cond_exception (exp_name, cond)
     char *exp_name;
     tree cond;
{
  /* jump around the exception if condition is NOT trrue. */
  if (cond)
    expand_start_cond (
      build1 (TRUTH_NOT_EXPR, boolean_type_node,
	chill_truthvalue_conversion (fold (cond))), 0);

  expand_cause_exception (exp_name);
  if (cond)
    expand_end_cond ();
}

/* If CONDITION is true, raise EXCEPTION (an IDENTIFIER_NODE);
   otherwise return EXPR. */

tree
check_expression (expr, condition, exception)
     tree expr, condition, exception;
{
  if (integer_zerop (condition))
    return expr;
  return build (COMPOUND_EXPR, TREE_TYPE (expr),
	   fold (build (TRUTH_ANDIF_EXPR, void_type_node, condition,
	     build_cause_exception (IDENTIFIER_POINTER (exception)))),
		expr);
}

/* Return EXPR, except if range_checking is on, return an expression
   that also checks that value >= low_limit && value <= hi_limit.
   If LO_LIMIT or HI_LIMIT is NULL_TREE, assume that test passes.  */

tree
check_range (expr, value, lo_limit, hi_limit)
     tree expr, value, lo_limit, hi_limit;
{
  if (lo_limit || hi_limit)
    {
      int old_inhibit_warnings = inhibit_warnings;
      tree lo_check, hi_check, check;

      /* This is a hack so that `shorten_compare' doesn't warn the
	 user about useless range checks that are too much work to
	 optimize away here.  */
      inhibit_warnings = 1;

      lo_check = lo_limit ? 
	fold (build_compare_discrete_expr (LT_EXPR, value, lo_limit)) :
	  boolean_false_node;   /* fake passing the check */

      hi_check = hi_limit ? 
	fold (build_compare_discrete_expr (GT_EXPR, value, hi_limit)) :
	  boolean_false_node;   /* fake passing the check */

      if (lo_check == boolean_false_node && hi_check == boolean_false_node)
	check = boolean_false_node;      /* fake passing the check */
      else if (lo_check == boolean_false_node)
	check = hi_check;
      else if (hi_check == boolean_false_node)
	check = lo_check;
      else
	check = fold (build (TRUTH_ORIF_EXPR, boolean_type_node,
			     lo_check, hi_check));

      inhibit_warnings = old_inhibit_warnings;

      if (!integer_zerop (check))
	{
	  if (current_function_decl == NULL_TREE)
	    {
	      if (TREE_CODE (check) == INTEGER_CST)
		error ("range failure (not inside function)");
	      else
		warning ("possible range failure (not inside function)");
	    }
	  else
	    {
	      if (TREE_CODE (check) == INTEGER_CST)
		warning ("expression will always cause RANGEFAIL");
	      if (range_checking)
		expr = check_expression (expr, check,
					 ridpointers[(int) RID_RANGEFAIL]);
	    }
	}
    }
  return expr;
}

/*
 * There are four conditions to generate a runtime check:
 *    1) assigning a longer INT to a shorter (signs irrelevant)
 *    2) assigning a signed to an unsigned
 *    3) assigning an unsigned to a signed of the same size.
 *    4) TYPE is a discrete subrange
 */
tree
chill_convert_for_assignment (type, expr, place)
     tree type, expr;
     char *place; /* location description for error messages */
{
  tree ttype = type;
  tree etype = TREE_TYPE (expr);

  if (type == NULL_TREE || TREE_CODE (type) == ERROR_MARK)
    return error_mark_node;
  if (expr == NULL_TREE || TREE_CODE (expr) == ERROR_MARK)
    return expr;

  if (! CH_COMPATIBLE (expr, type))
    {
      error ("incompatible modes in %s", place);
      return error_mark_node;
    }

  if (TREE_CODE (type) == REFERENCE_TYPE)
    ttype = TREE_TYPE (ttype);
  if (etype && TREE_CODE (etype) == REFERENCE_TYPE)
    etype = TREE_TYPE (etype);

  if (CH_STRING_TYPE_P (ttype) && etype && CH_STRING_TYPE_P (etype))
    {
      tree cond = string_assignment_condition (ttype, expr);
      if (TREE_CODE (cond) == INTEGER_CST)
	{
	  if (tree_int_cst_equal (cond, integer_zero_node))
	    {
	      error ("bad string length in %s", place);
	      return error_mark_node;
	    }
	  /* Otherwise, the condition is always true, so no runtime test. */
	}
      else if (range_checking)
	{
	  expr = build (COMPOUND_EXPR, TREE_TYPE (expr),
			fold (build (TRUTH_ORIF_EXPR, void_type_node, cond,
				     build_cause_exception ("RANGE_FAIL"))),
			expr);
	}
    }

  if (range_checking 
      && discrete_type_p (ttype) 
      && etype != NULL_TREE
      && discrete_type_p (etype))
    {
      int cond1 = tree_int_cst_lt (TYPE_SIZE (ttype),
				   TYPE_SIZE (etype));
      int cond2 = TREE_UNSIGNED (ttype) 
	          && (! TREE_UNSIGNED (etype));
      int cond3 = (! TREE_UNSIGNED (type))
	          && TREE_UNSIGNED (etype) 
		  && tree_int_cst_equal (TYPE_SIZE (ttype),
					 TYPE_SIZE (etype));
      int cond4 = TREE_TYPE (ttype) 
	          && discrete_type_p (TREE_TYPE (ttype));

      if (cond1 || cond2 || cond3 || cond4)
	{
	  tree type_min = TYPE_MIN_VALUE (ttype);
	  tree type_max = TYPE_MAX_VALUE (ttype);

	  /* expr = save_expr (expr); ?? */
	  if (expr && type_min && type_max)
	    expr = check_range (expr, expr, type_min, type_max);
	}
    }
  return convert (type, expr);
}

tree
adjust_return_value (expr, res_ret_flag)
     tree expr;
     int  res_ret_flag;
{
  tree type = TREE_TYPE (TREE_TYPE (current_function_decl));

  if (TREE_CODE (type) == REFERENCE_TYPE)
    {
      if (CH_LOCATION_P (expr))
	{
	  if (! CH_READ_COMPATIBLE (TREE_TYPE (type), 
				    TREE_TYPE (expr)))
	    {
	      error ("mode mismatch in %s expression", 
		     res_ret_flag ? "result" : "return");
	      return error_mark_node;
	    }
	  return convert (type, expr);
	}
      else
	{
	  tree temp;

	  if (! CH_COMPATIBLE (expr, TREE_TYPE (type)))
	    {
	      error ("mode mismatch in %s expression", 
		     res_ret_flag ? "result" : "return");
	      return error_mark_node;
	    }
	  temp = decl_temp1 (get_unique_identifier ("RETN"),
			     TREE_TYPE (type), 0, expr, 0, 0);
	  mark_addressable (temp);
	  return build1 (ADDR_EXPR, type, temp);
	}
    }
  else if (! CH_COMPATIBLE (expr, type))
    {
      error ("mode mismatch in %s expression", 
	     res_ret_flag ? "result" : "return");
      return error_mark_node;
    }
  return convert (type, expr);
}

void
chill_expand_result (expr)
     tree expr;
{
  tree type;

  CH_FUNCTION_SETS_RESULT (current_function_decl) = 1;

  if (pass == 1)
    return;

  if (expr == NULL_TREE || TREE_CODE (expr) == ERROR_MARK)
    return;

  if (chill_at_module_level || global_bindings_p ())
    error ("RESULT not allowed outside a PROC");

  result_never_set = 0;

  if (chill_result_decl == NULL_TREE)
    {
      error ("RESULT action in PROC with no declared result");
      return;
    }
  type = TREE_TYPE (chill_result_decl);

  expr = adjust_return_value (expr, 1);

  expand_expr_stmt (build_chill_modify_expr (chill_result_decl, expr));
}

/*
 * error if EXPR not NULL and procedure doesn't
 * have a return type; 
 * warning if EXPR NULL,
 * procedure *has* a return type, and a previous
 * RESULT actions hasn't saved a return value.
 */
void
chill_expand_return (expr, implicit)
     tree expr;
     int implicit; /* 1 if an implicit return at end of function. */
{
  tree valtype;

  if (expr != NULL_TREE && TREE_CODE (expr) == ERROR_MARK)
    return;
  if (chill_at_module_level || global_bindings_p ())
    {
      error ("RETURN not allowed outside PROC");
      return;
    }

  if (pass == 1)
    return;

  result_never_set = 0;

  valtype = TREE_TYPE (TREE_TYPE (current_function_decl));
  if (TREE_CODE (valtype) == VOID_TYPE)
    {
      if (expr != NULL_TREE)
	error ("RETURN with a value, in PROC returning void");
      expand_null_return ();
    }
  else
    {
      if (expr == NULL_TREE)
	{
	  if (!CH_FUNCTION_SETS_RESULT (current_function_decl)
	      && !implicit)
	    warning ("RETURN with no value, in function with no RESULT");
	  expr = chill_result_decl;
	}
      else
	expr = adjust_return_value (expr, 0);
      expr = build (MODIFY_EXPR, valtype,
		    DECL_RESULT (current_function_decl),
		    expr);
      TREE_SIDE_EFFECTS (expr) = 1;
      expand_return (expr);
    }
}

void
lookup_and_expand_goto (name)
     tree name;
{
  if (!ignoring)
    {
      tree decl = lookup_name (name);
      if (decl == NULL || TREE_CODE (decl) != LABEL_DECL)
	error ("no label named `%s'", IDENTIFIER_POINTER (name));
      else if (DECL_CONTEXT (decl) != current_function_decl)
	error ("cannot GOTO label `%s' outside current function",
	       IDENTIFIER_POINTER (name));
      else
	{
	  TREE_USED (decl) = 1;
	  expand_goto_except_cleanup (DECL_ACTION_NESTING_LEVEL (decl));
	  expand_goto (decl);
	}
    }
}

void
lookup_and_handle_exit (name)
     tree name;
{
  if (!ignoring)
    {
      tree label = munge_exit_label (name);
      tree decl = lookup_name (label);
      if (decl == NULL || TREE_CODE (decl) != LABEL_DECL)
	{
	  if (pass == 2)
	    error ("no EXITable label named `%s'", IDENTIFIER_POINTER (name));
	}
      else if (DECL_CONTEXT (decl) != current_function_decl)
	{
	  if (pass == 2)
	    error ("cannot EXIT label `%s' outside current function",
		   IDENTIFIER_POINTER (name));
	}
      else
	{
	  TREE_USED (decl) = 1;
	  expand_goto_except_cleanup (DECL_ACTION_NESTING_LEVEL (decl));
	  expand_goto (decl);
	}
    }
}

void
chill_handle_case_default ()
{
  tree duplicate;
  register tree label = build_decl (LABEL_DECL, NULL_TREE, 
				    NULL_TREE);
  int success = pushcase (NULL_TREE, 0, label, &duplicate);

  if (success == 1)
    error ("ELSE label not within a CASE statement");
#if 0
  else if (success == 2)
    {
      error ("multiple default labels found in a CASE statement"); 
      error_with_decl (duplicate, "this is the first ELSE label");
    }
#endif
}

/* Handle cases label such as (I:J):  or (modename): */

static void
chill_handle_case_label_range (min_value, max_value)
     tree min_value, max_value;
{
  register tree label = build_decl (LABEL_DECL, NULL_TREE, NULL_TREE);
  if (TREE_CODE (min_value) != ERROR_MARK
      && TREE_CODE (max_value) != ERROR_MARK)
    {
      tree duplicate;
      int success = pushcase_range (min_value, max_value,
				    convert, label, &duplicate);
      if (success == 1)
	error ("label found outside of CASE statement");
      else if (success == 2)
	{
	  error ("duplicate CASE value");
	  error_with_decl (duplicate, "this is the first entry for that value");
	}
      else if (success == 3)
	error ("CASE value out of range");
      else if (success == 4)
	error ("empty range");
      else if (success == 5)
	error ("label within scope of cleanup or variable array");
    }
}

void
chill_handle_case_label (label_value)
     tree label_value;
{
  if (label_value == NULL_TREE 
      || TREE_CODE (label_value) == ERROR_MARK)
    return;
  if (TREE_CODE (label_value) == RANGE_EXPR)
    {
      if (TREE_OPERAND (label_value, 0) == NULL_TREE)
	chill_handle_case_default ();  /* i.e. (ELSE): or (*): */
      else
	chill_handle_case_label_range (TREE_OPERAND (label_value, 0),
				       TREE_OPERAND (label_value, 1));
    }
  else if (TREE_CODE (label_value) == TYPE_DECL)
    {
      tree type = TREE_TYPE (label_value);
      if (! discrete_type_p (type))
	error ("mode in label is not discrete");
      else
	chill_handle_case_label_range (TYPE_MIN_VALUE (type),
				       TYPE_MAX_VALUE (type));
    }
  else
    { register tree value = check_case_value (label_value);
      register tree label
	= build_decl (LABEL_DECL, NULL_TREE, NULL_TREE);

      if (TREE_CODE (value) != ERROR_MARK)
	{
	  tree duplicate;
	  int success = pushcase (value, convert, label, &duplicate);
	  if (success == 1)
	    error ("label not within a CASE statement");
	  else if (success == 2)
	    {
	      error ("duplicate case value");
	      error_with_decl (duplicate, 
			       "this is the first entry for that value");
	    }
	  else if (success == 3)
	    error ("CASE value out of range");
	  else if (success == 4)
	    error ("empty range");
	  else if (success == 5)
	    error ("label within scope of cleanup or variable array");
        }
    }
}

#define BITARRAY_TEST(ARRAY, INDEX) \
  ((ARRAY)[(unsigned)(INDEX) / HOST_BITS_PER_CHAR]\
			  & (1 << ((unsigned)(INDEX) % HOST_BITS_PER_CHAR)))
#define BITARRAY_SET(ARRAY, INDEX) \
  ((ARRAY)[(unsigned)(INDEX) / HOST_BITS_PER_CHAR]\
			  |= 1 << ((unsigned)(INDEX) % HOST_BITS_PER_CHAR))

extern HOST_WIDE_INT all_cases_count PROTO((tree, int*));
extern void mark_seen_cases PROTO((tree, unsigned char*, long, int));

/* CASES_SEEN is a set (bitarray) of length COUNT.
   For each element that is zero, print an error message,
   assume the element have the given TYPE. */

static void
print_missing_cases (type, cases_seen, count)
     tree type;
     unsigned char *cases_seen;
     long count;
{
  long i;
  for (i = 0;  i < count; i++)
    {
      if (BITARRAY_TEST(cases_seen, i) == 0)
	{
	  char buf[20];
	  long x = i;
	  long j;
	  tree t = type;
	  char *err_val_name = "???";
	  if (TYPE_MIN_VALUE (t)
	      && TREE_CODE (TYPE_MIN_VALUE (t)) == INTEGER_CST)
	    x += TREE_INT_CST_LOW (TYPE_MIN_VALUE (t));
	  while (TREE_TYPE (t) != NULL_TREE)
	    t = TREE_TYPE (t);
	  switch (TREE_CODE (t))
	    {
	      tree v;
	    case BOOLEAN_TYPE:
	      err_val_name = x ? "TRUE" : "FALSE";
	      break;
	    case CHAR_TYPE:
	      if ((x >= ' ' && x < 127) && x != '\'' && x != '^')
		sprintf (buf, "'%c'", x);
	      else
		sprintf (buf, "'^(%d)'", x);
	      err_val_name = buf;
	      j = i;
	      while (j < count && !BITARRAY_TEST(cases_seen, j))
		j++;
	      if (j > i + 1)
		{
		  long y = x+j-i-1;
		  err_val_name += strlen (err_val_name);
		  if ((y >= ' ' && y < 127) && y != '\'' && y != '^')
		    sprintf (err_val_name, "%s:'%c'", buf, x);
		  else
		    sprintf (err_val_name, "%s:'^(%d)'", buf, x);
		  i = j - 1;      
		}
	      break;
	    case ENUMERAL_TYPE:
	      for (v = TYPE_VALUES (t);  v && x;  v = TREE_CHAIN (v))
		x--;
	      if (v)
		err_val_name = IDENTIFIER_POINTER (TREE_PURPOSE (v));
	      break;
	    default:
	      j = i;
	      while (j < count && !BITARRAY_TEST(cases_seen, j))
		j++;
	      if (j == i + 1)
		sprintf (buf, "%d", x);
	      else
		sprintf (buf, "%d:%d", x, x+j-i-1);
	      i = j - 1;      
	      err_val_name = buf;
	      break;
	    }
	  error ("incomplete CASE - %s not handled", err_val_name);
	}
    }
}

void
check_missing_cases (type)
     tree type;
{
  int is_sparse;
  /* For each possible selector value. a one iff it has been matched
     by a case value alternative. */
  unsigned char *cases_seen;
  /* The number of possible selector values. */
  HOST_WIDE_INT size = all_cases_count (type, &is_sparse);
  long bytes_needed = (size+HOST_BITS_PER_CHAR)/HOST_BITS_PER_CHAR;

  if (size == -1)
    warning ("CASE selector with variable range");
  else if (size < 0 || size > 600000
	   /* We deliberately use malloc here - not xmalloc. */
	   || (cases_seen = (char*) malloc (bytes_needed)) == NULL)
    warning ("too many cases to do CASE completeness testing");
  else
    {
      bzero (cases_seen, bytes_needed);
      mark_seen_cases (type, cases_seen, size, is_sparse);
      print_missing_cases (type, cases_seen, size);
      free (cases_seen);
    }
}

/*
 * We build an expression tree here because, in many contexts,
 * we don't know the type of result that's desired.  By the
 * time we get to expanding the tree, we do know.
 */
tree
build_chill_case_expr (exprlist, casealtlist_expr,
		       optelsecase_expr)
     tree exprlist, casealtlist_expr, optelsecase_expr;
{
  return build (CASE_EXPR, NULL_TREE, exprlist,
		nreverse (optelsecase_expr ?
			  tree_cons (NULL_TREE, 
				     optelsecase_expr,
				     casealtlist_expr) :
			  casealtlist_expr));
}

/* The index expression has already been validated.  We now need
 * to:
 *    type-check the assignment-to-be
 *    if (index - domain_min_value) * element_size == length_fld)
 *      length_fld += element_size
 *    perform the assignment.
 * LHS is a VARY_ARRAY_REF expression.
 */
static void
expand_assignment_to_varying_array (lhs, rhs)
     tree lhs, rhs;
{
  tree array          = TREE_OPERAND (lhs, 0);
  tree index          = TREE_OPERAND (lhs, 1);
  tree base_array     = TREE_CHAIN (TYPE_FIELDS (TREE_TYPE (array)));  
  tree element_size   = size_in_bytes (TREE_TYPE (TREE_TYPE (base_array)));
  tree min_domain_val = TYPE_MIN_VALUE (TYPE_DOMAIN (TREE_TYPE (base_array)));
  tree length_fld     = stabilize_reference (
		          convert (integer_type_node,
                            build_component_ref (array, var_length_id)));
  /* FIXME: we have this same code somewhere else. */
  tree cond = build (EQ_EXPR, boolean_type_node,
		     length_fld,
		     fold (size_binop (MULT_EXPR,
				       size_binop (MINUS_EXPR,
						   index,
						   min_domain_val),
				       element_size)));
  expand_start_cond (chill_truthvalue_conversion (fold (cond)), 0);
  expand_expr_stmt
    (build_chill_modify_expr (length_fld,
			      build_chill_binary_op (PLUS_EXPR,
						     length_fld,
						     element_size)));
  expand_end_cond ();
  array = build_component_ref (array, var_data_id);
  chill_expand_assignment (build_chill_array_ref (array, index),
			   NOP_EXPR, rhs);
}

/*
 * called to handle a LHS and RHS which are arrays.
 * Here's the assignment:
 *              LHS := RHS;   -- where 'LHS' means 'left-hand-side' etc.
 * Also called to convert procedure parameters.
 */
static void
chill_expand_array_assignment (lhs, rhs)
     tree lhs, rhs;
{
  int lhs_varying = chill_varying_type_p (TREE_TYPE (lhs));
  int rhs_varying = chill_varying_type_p (TREE_TYPE (rhs));

  /* point at actual RHS data's type */
  tree rhs_data_type = rhs_varying ? 
                       CH_VARYING_ARRAY_TYPE (TREE_TYPE (rhs)) :
		       TREE_TYPE (rhs);
  {
    /* point at actual LHS data's type */
    tree lhs_data_type = lhs_varying ? 
                         CH_VARYING_ARRAY_TYPE (TREE_TYPE (lhs)) :
			 TREE_TYPE (lhs);

    int lhs_bytes = int_size_in_bytes (lhs_data_type);
    int rhs_bytes = int_size_in_bytes (rhs_data_type);

    /* if both sides not varying, and sizes not dynamically 
       computed, sizes must *match* */
    if (! lhs_varying && ! rhs_varying && lhs_bytes != rhs_bytes
	&& lhs_bytes > 0 && rhs_bytes > 0)
      {
	error ("string lengths not equal");
	return;
      }
    /* Must have enough space on LHS for static size of RHS */
    
    if (lhs_bytes > 0 && rhs_bytes > 0 
	&& lhs_bytes < rhs_bytes)
      {
	if (rhs_varying)
	  {
	    /* FIXME: generate runtime test for room */
	    ;
	  }
	else
	  {
	    error ("can't do ARRAY assignment - too large");
	    return;
	  }
      }
  }
  
  /* now we know the RHS will fit in LHS, build trees for the
     emit_block_move parameters */

  if (lhs_varying)
    {
      rhs = convert (TREE_TYPE (lhs), rhs);
      expand_expr_stmt (
        build_chill_modify_expr (lhs, convert (TREE_TYPE (lhs), rhs)));
      return;
    }
  else if (rhs_varying)
    rhs = build_component_ref (rhs, var_data_id);

  if (! mark_addressable (rhs))
    {
      error ("rhs of array assignment is not addressable");
      return;
    }

  expand_expr_stmt (
    build_chill_function_call (lookup_name (get_identifier ("memmove")),
      tree_cons (NULL_TREE, build_chill_addr_expr (lhs, "lhs of array assignment"),
        tree_cons (NULL_TREE, build1 (ADDR_EXPR, const_ptr_type_node, rhs),
          tree_cons (NULL_TREE, size_in_bytes (rhs_data_type), 
		     NULL_TREE)))));
#if 0
  store_expr (rhs,
    expand_expr (lhs, NULL_RTX, BLKmode, EXPAND_NORMAL), 0);
#endif
}

/* This is called with the assumption that RHS has been stabilized.  
   It has one purpose:  to iterate through the CHILL list of LHS's */
void
expand_assignment_action (loclist, modifycode, rhs)
     tree loclist;
     enum chill_tree_code modifycode;
     tree rhs;
{
  if (loclist == NULL_TREE || TREE_CODE (loclist) == ERROR_MARK
      || rhs == NULL_TREE  || TREE_CODE (rhs) == ERROR_MARK)
    return;

  if (TREE_CHAIN (loclist) != NULL_TREE)
    { /* Multiple assignment */
      tree target;
      if (TREE_TYPE (rhs) != NULL_TREE)
	rhs = save_expr (rhs);
      else if (TREE_CODE (rhs) == CONSTRUCTOR)
	error ("type of tuple cannot be implicit in multiple assignent");
      else if (TREE_CODE (rhs) == CASE_EXPR || TREE_CODE (rhs) == COND_EXPR)
	error ("conditional expression cannot be used in multiple assignent");
      else
	error ("internal error - unknown type in multiple assignment");

      if (modifycode != NOP_EXPR)
	{
	  error ("no operator allowed in multiple assignment,");
	  modifycode = NOP_EXPR;
	}

      for (target = TREE_CHAIN (loclist); target; target = TREE_CHAIN (target))
	{
	  if (!CH_EQUIVALENT (TREE_TYPE (TREE_VALUE (target)),
			      TREE_TYPE (TREE_VALUE (loclist))))
	    {
	      error
		("location modes in multiple assignment are not equivalent");
	      break;
	    }
	}
    }
  for ( ; loclist != NULL_TREE; loclist = TREE_CHAIN (loclist))
    chill_expand_assignment (TREE_VALUE (loclist), modifycode, rhs);
}

void
chill_expand_assignment (lhs, modifycode, rhs)
     tree lhs;
     enum chill_tree_code modifycode;
     tree rhs;
{
  if (TREE_CODE (lhs) == ERROR_MARK)
    return;

  /* errors for assignment to BUFFER, EVENT locations.
     what about SIGNALs? FIXME: Need similar test in
     build_chill_function_call. */
  if (TREE_CODE (lhs) == IDENTIFIER_NODE)
    {
      tree decl = lookup_name (lhs);
      if (decl)
	{
	  tree type = TREE_TYPE (decl);
	  if (CH_IS_BUFFER_MODE (type) || CH_IS_EVENT_MODE (type))
	    {
	      error ("You may not assign a value to a BUFFER or EVENT location");
	      return;
	    }
	}
    }

  if (TREE_READONLY (TREE_TYPE (lhs)))
    {
      error ("can't assign value to READonly location");
      return;
    }

  if (TREE_CODE (TREE_TYPE (lhs)) == REFERENCE_TYPE)
    lhs = build1 (INDIRECT_REF, TREE_TYPE (TREE_TYPE (lhs)), lhs);

  /* If a binary op has been requested, combine the old LHS value with
     the RHS producing the value we should actually store into the LHS. */

  if (modifycode != NOP_EXPR)
    {
      lhs = stabilize_reference (lhs);
      /* This is to handle border-line cases such
	 as: LHS OR := [I].  This seems to be permitted
	 by the letter of Z.200, though it violates
	 its spirit, since LHS:=LHS OR [I] is
	 *not* legal. */
      if (TREE_TYPE (rhs) == NULL_TREE)
	rhs = convert (TREE_TYPE (lhs), rhs);
      rhs = build_chill_binary_op (modifycode, lhs, rhs);
    }

  rhs = chill_convert_for_assignment (TREE_TYPE (lhs), rhs, "assignment");

#if 0
  if (TREE_CODE (rhs) == VARY_ARRAY_REF)
    {
      tree array     = TREE_OPERAND (rhs, 0);
      tree index = TREE_OPERAND (rhs, 1);
#if 0
      if (chill_varying_type_p (TREE_TYPE (array)))
	array = fold (build_component_ref (array, var_data_id));
#endif
      rhs = build_chill_array_ref (array, index);
	}
#endif
  /* handle the LENGTH (vary_array) := expr action */
  {
    tree loc = lhs;
    if (TREE_CODE (loc) == NOP_EXPR)
      loc = TREE_OPERAND (loc, 0);
    if (TREE_CODE (loc) == COMPONENT_REF
	&& chill_varying_type_p (TREE_TYPE (TREE_OPERAND (loc, 0)))
	&& DECL_NAME (TREE_OPERAND (loc, 1)) == var_length_id)
      expand_varying_length_assignment (TREE_OPERAND (loc, 0), rhs);
  }
  /* handle array assignments specially, including VARYING ones */
  if (TREE_CODE (lhs) == VARY_ARRAY_REF
      && chill_varying_type_p (TREE_TYPE (TREE_OPERAND (lhs, 0))))
    {
      expand_assignment_to_varying_array (lhs, rhs);
      return;
    }

  if (TREE_CODE (lhs) == SLICE_EXPR)
    {
      tree func = lookup_name (get_identifier ("__pscpy"));
      tree dst = TREE_OPERAND (lhs, 0);
      tree dst_offset = TREE_OPERAND (lhs, 1);
      tree length = TREE_OPERAND (lhs, 2);
      tree src, src_offset;
      if (TREE_CODE (rhs) == SLICE_EXPR)
	{
	  src = TREE_OPERAND (rhs, 0);
	  /* Should check that the TREE_OPERAND (src, 0) is
	     the same as length and powerserlen (src).  FIXME */
	  src_offset = TREE_OPERAND (rhs, 1);
	}
      else
	{
	  src = rhs;
	  src_offset = integer_zero_node;
	}
      expand_expr_stmt (build_chill_function_call (func,
	tree_cons (NULL_TREE, force_addr_of (dst),
	  tree_cons (NULL_TREE, powersetlen (dst),
	    tree_cons (NULL_TREE, dst_offset,
	      tree_cons (NULL_TREE, force_addr_of (src),
		tree_cons (NULL_TREE, powersetlen (src),
		  tree_cons (NULL_TREE, src_offset,
		    tree_cons (NULL_TREE, length, NULL_TREE)))))))));
    }

  if ((TREE_CODE (TREE_TYPE (lhs)) == ARRAY_TYPE &&
       (TREE_TYPE (rhs) && TREE_CODE (TREE_TYPE (rhs)) == ARRAY_TYPE)) ||
      chill_varying_type_p (TREE_TYPE (lhs)) || 
      chill_varying_type_p (TREE_TYPE (rhs)))
    {
      chill_expand_array_assignment (lhs, rhs);
      return;
    }

  if (TREE_CODE (lhs) == SET_IN_EXPR)
    {
      tree from_pos = save_expr (TREE_OPERAND (lhs, 0));
      tree set = TREE_OPERAND (lhs, 1);
      tree domain = TYPE_DOMAIN (TREE_TYPE (set));
      tree set_length = size_binop (PLUS_EXPR,
				    size_binop (MINUS_EXPR,
						TYPE_MAX_VALUE (domain),
						TYPE_MIN_VALUE (domain)),
				    integer_one_node);
      tree filename = force_addr_of (get_chill_filename());
      
      if (TREE_CODE (TREE_TYPE (lhs)) != BOOLEAN_TYPE)
	sorry("bitstring slice");
      expand_expr_stmt (
	build_chill_function_call (lookup_name (
	  get_identifier ("__setbitpowerset")),
	      tree_cons (NULL_TREE, build_chill_addr_expr (set, "powerset"),
		  tree_cons (NULL_TREE, set_length,
		    tree_cons (NULL_TREE, TYPE_MIN_VALUE (domain),
		      tree_cons (NULL_TREE, from_pos,
			tree_cons (NULL_TREE, rhs,
			  tree_cons (NULL_TREE, filename,
			    tree_cons (NULL_TREE, get_chill_linenumber(),
  			      NULL_TREE)))))))));
      return;
    }

  /* The following is probably superceded by the
     above code for SET_IN_EXPR. FIXME! */
  if (TREE_CODE (lhs) == BIT_FIELD_REF)
    {
      tree set = TREE_OPERAND (lhs, 0);
      tree numbits = TREE_OPERAND (lhs, 1);
      tree from_pos = save_expr (TREE_OPERAND (lhs, 2));
      tree domain = TYPE_DOMAIN (TREE_TYPE (set));
      tree set_length = size_binop (PLUS_EXPR,
				    size_binop (MINUS_EXPR,
						TYPE_MAX_VALUE (domain),
						TYPE_MIN_VALUE (domain)),
				    integer_one_node);
      tree filename = force_addr_of (get_chill_filename());
      tree to_pos;
      switch (TREE_CODE (TREE_TYPE (rhs)))
	{
	case SET_TYPE:
	  to_pos = size_binop (MINUS_EXPR,
			       size_binop (PLUS_EXPR, from_pos, numbits),
			       integer_one_node);
	  break;
	case BOOLEAN_TYPE:
	  to_pos = from_pos;
	  break;
	default:
	  abort ();
	}
      
      if (TREE_CODE (TREE_TYPE (lhs)) != BOOLEAN_TYPE)
	sorry("bitstring slice");
      expand_expr_stmt (
	  build_chill_function_call( lookup_name (
	      get_identifier ("__setbitpowerset")),
		tree_cons (NULL_TREE, build_chill_addr_expr (set, "powerset"),
		  tree_cons (NULL_TREE, set_length,
		    tree_cons (NULL_TREE, TYPE_MIN_VALUE (domain),
		      tree_cons (NULL_TREE, from_pos,
			tree_cons (NULL_TREE, rhs,
			  tree_cons (NULL_TREE, filename,
			    tree_cons (NULL_TREE, get_chill_linenumber(),
	  		      NULL_TREE)))))))));
      return;
    }

  expand_expr_stmt (build_chill_modify_expr (lhs, rhs));
}

/* Also assumes that rhs has been stabilized */
void
expand_varying_length_assignment (lhs, rhs)
     tree lhs, rhs;
{
  tree base_array, min_domain_val;
  tree result, cond_list = NULL_TREE;

  pedwarn ("LENGTH on left-hand-side is non-portable");
      
  if (! CH_LOCATION_P (lhs))
    {
      error ("Can only set LENGTH of array location");
      return;
    }

  /* cause a RANGE exception if rhs would cause a 'hole' in the array. */
  if (valid_array_index_p (lhs, rhs, &cond_list, 1) == NULL_TREE)
    {
      error ("new array length too large");
      return;
    }

  base_array     = CH_VARYING_ARRAY_TYPE (TREE_TYPE (lhs));
  min_domain_val = TYPE_MIN_VALUE (TYPE_DOMAIN (base_array));

  result = build_chill_modify_expr (
	     build_component_ref (lhs, var_length_id),
               size_binop (MINUS_EXPR, rhs, min_domain_val));

  cond_list = tree_cons (NULL_TREE, result, cond_list);
  expand_expr_stmt (build_chill_compound_expr (nreverse (cond_list)));
}

void
push_action ()
{
  if (global_bindings_p ())
      chill_start_module_code (); 
  push_handler ();
  if (ignoring)
    return;
  emit_line_note (input_filename, lineno);
}
