/* Implement grant-file output & seize-file input for CHILL.
   Copyright (C) 1992, 93, 1994 Free Software Foundation, Inc.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include "config.h"
#include "tree.h"
#include "ch-tree.h"
#include "lex.h"
#include "flags.h"
#include "actions.h"
#include "input.h"
#include "errno.h"
#include "rtl.h"
#include "tasking.h"

/* Disable possible macro over-rides, so the externs parse
   portably. */
#undef strchr
#undef strrchr

#define APPEND(X,Y) X = append (X, Y)
#define FREE(x) strfree (x)
#define ALLOCAMOUNT	10000
/* may be we can handle this in a more exciting way,
   but this also should work for the moment */

extern void assemble_constructor PROTO((char *));
extern void assemble_name PROTO((FILE *, char *));
extern void error PROTO((char *, ...));
extern tree tasking_list;
extern void tasking_registry PROTO((void));
extern void tasking_setup PROTO((void));
extern void build_enum_tables PROTO((void));
extern tree process_type;
extern void warning PROTO((char *, ...));
extern tree get_file_function_name PROTO((int));
extern char *asm_file_name;
extern char *dump_base_name;

/* forward declarations */

/* variable indicates compilation at module level */
int chill_at_module_level = 0;

/* this variable holds the module name */
tree chill_module_name = NULL_TREE;

/* variable holds the name of the initializer of this module, if non-NULL */
tree module_function_name = NULL_TREE;

/* mark that a SPEC MODULE was generated */
static int spec_module_generated = 0;

/* define version strings */
extern char *gnuchill_version;
extern char *version_string;

/* define a faster string handling */
typedef struct
{
  char	*str;
  int		len;
  int		allocated;
} MYSTRING;

/* structure used for handling multiple grant files */
char	*grant_file_name;
MYSTRING	*gstring = NULL;

static MYSTRING *decode_decl             PROTO((tree, tree));
static MYSTRING *decode_constant         PROTO((tree));
static void      grant_one_decl          PROTO((tree));
static MYSTRING *get_type                PROTO((tree, tree));
static MYSTRING *decode_mode             PROTO((tree, tree));

/* list of the VAR_DECLs of the module initializer entries */
static tree      module_init_list = NULL_TREE;

static MYSTRING *
newstring (str)
    char	*str;
{
    MYSTRING	*tmp = (MYSTRING *) xmalloc (sizeof (MYSTRING));
    unsigned	len = strlen (str);
    
    tmp->allocated = len + ALLOCAMOUNT;
    tmp->str = xmalloc ((unsigned)tmp->allocated);
    strcpy (tmp->str, str);
    tmp->len = len;
    return (tmp);
}

static void
strfree (str)
    MYSTRING	*str;
{
    free (str->str);
    free (str);
}

static MYSTRING *
append (inout, in)
    MYSTRING	*inout;
    char	*in;
{
    int		inlen = strlen (in);
    
    if ((inout->len + inlen) >= inout->allocated)
	inout->str = xrealloc (inout->str, inout->allocated += ALLOCAMOUNT);
    strcpy (inout->str + inout->len, in);
    inout->len += inlen;
    return (inout);
}

void
grant_use_seizefile (seize_filename)
     char *seize_filename;
{
  APPEND (gstring, "<> USE_SEIZE_FILE \"");
  APPEND (gstring, seize_filename);
  APPEND (gstring, "\" <>\n");
}

MYSTRING *
grant_array_type (type)
     tree type;
{
  MYSTRING	*result = newstring ("");
  MYSTRING	*mode_string;
  int            varying = 0;

  if (chill_varying_type_p (type))
    {
      varying = 1;
      type = CH_VARYING_ARRAY_TYPE (type);
    }
  if (CH_STRING_TYPE_P (type))
    {
      tree fields = TYPE_DOMAIN (type);
      tree maxval = TYPE_MAX_VALUE (fields);

      if (TREE_CODE (TREE_TYPE (type)) == CHAR_TYPE)
	APPEND (result, "CHARS (");
      else
	APPEND (result, "BOOLS (");
      if (TREE_CODE (maxval) == INTEGER_CST)
	{
	  char	wrk[20];
	  sprintf (wrk, "%d", TREE_INT_CST_LOW (maxval) + 1);
	  APPEND (result, wrk);
	}
      else if (TREE_CODE (maxval) == MINUS_EXPR
	       && TREE_OPERAND (maxval, 1) == integer_one_node)
	{
	  mode_string = decode_constant (TREE_OPERAND (maxval, 0));
	  APPEND (result, mode_string->str);
	  FREE (mode_string);
	}
      else
	{
	  mode_string = decode_constant (maxval);
	  APPEND (result, mode_string->str);
	  FREE (mode_string);
	  APPEND (result, "+1");
	}
      APPEND (result, ")");
      if (varying)
	APPEND (result, " VARYING");
      return result;
    }

  APPEND (result, "ARRAY (");
  if (TREE_CODE (TYPE_DOMAIN (type)) == INTEGER_TYPE
     && TREE_TYPE (TYPE_DOMAIN (type)) == ridpointers[(int) RID_RANGE])
    {
      mode_string = decode_constant (TYPE_MIN_VALUE (TYPE_DOMAIN (type)));
      APPEND (result, mode_string->str);
      FREE (mode_string);
      
      APPEND (result, ":");
      mode_string = decode_constant (TYPE_MAX_VALUE (TYPE_DOMAIN (type)));
      APPEND (result, mode_string->str);
      FREE (mode_string);
    }
  else
    {
      mode_string = decode_mode (TYPE_DOMAIN (type), NULL_TREE);
      APPEND (result, mode_string->str);
      FREE (mode_string);
    }
  APPEND (result, ") ");
  if (varying)
    APPEND (result, "VARYING ");

  mode_string = get_type (TREE_TYPE (type), NULL_TREE);
  APPEND (result, mode_string->str);
  FREE (mode_string);

  return result;
}

static MYSTRING *
get_tag_value (val)
    tree	val;
{
  MYSTRING	*result;
    
  if (TREE_CODE (val) == CONST_DECL && DECL_NAME (val))
    {
      result = newstring ("");
      APPEND (result, IDENTIFIER_POINTER (DECL_NAME (val)));
    }
  else if (TREE_CODE (val) == CONST_DECL)
    {
      /* it's a synonym -- get the value */
      result = decode_constant (DECL_INITIAL (val));
    }
  else
    {
      result = decode_constant (val);
    }
  return (result);
}

static MYSTRING *
print_enumeral (type)
     tree type;
{
  MYSTRING	*result = newstring ("");
  tree	fields;

#if 0
  if (TYPE_LANG_SPECIFIC (type) == NULL)
#endif
    {
      
      APPEND (result, "SET (");
      for (fields = TYPE_VALUES (type);
	   fields != NULL_TREE;
	   fields = TREE_CHAIN (fields))
	{
	  if (TREE_PURPOSE (fields) == NULL_TREE)
	    APPEND (result, "*");
	  else
	    {
	      tree decl = TREE_VALUE (fields);
	      APPEND (result, IDENTIFIER_POINTER (TREE_PURPOSE (fields)));
	      if (TREE_CODE (decl) == CONST_DECL && DECL_INITIAL (decl))
		{
		  MYSTRING *val_string = decode_constant (DECL_INITIAL (decl));
		  APPEND (result, " = ");
		  APPEND (result, val_string->str);
		  FREE (val_string);
		}
	    }
	  if (TREE_CHAIN (fields) != NULL_TREE)
	    APPEND (result, ",\n     ");
	}
      APPEND (result, ")");
    }
  return result;
}

static MYSTRING *
print_integer_type (type, forbid)
     tree type;
     tree forbid;
{
  MYSTRING *result = newstring ("");
  MYSTRING *mode_string;
  char     *name_ptr;
  tree	    base_type;

  if (TREE_TYPE (type))
    {
      mode_string = decode_mode (TREE_TYPE (type), forbid);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      
      APPEND (result, "(");
      mode_string = decode_constant (TYPE_MIN_VALUE (type));
      APPEND (result, mode_string->str);
      FREE (mode_string);
      
      APPEND (result, ":");
      mode_string = decode_constant (TYPE_MAX_VALUE (type));
      APPEND (result, mode_string->str);
      FREE (mode_string);
      
      APPEND (result, ")");
      return result;
    }
  /* We test TYPE_MAIN_VARIANT because pushdecl often builds
     a copy of a built-in type node, which is logically id-
     entical but has a different address, and the same
     TYPE_MAIN_VARIANT. */

  base_type = TREE_TYPE (type) ? TREE_TYPE (type) : type;

  if (TREE_UNSIGNED (base_type))
    {
      if (base_type == chill_unsigned_type_node
	  || TYPE_MAIN_VARIANT(base_type) ==
	     TYPE_MAIN_VARIANT (chill_unsigned_type_node))
	name_ptr = "UINT";
      else if (base_type == long_integer_type_node
	       || TYPE_MAIN_VARIANT(base_type) ==
	          TYPE_MAIN_VARIANT (long_unsigned_type_node))
	name_ptr = "ULONG";
      else if (type == unsigned_char_type_node
	       || TYPE_MAIN_VARIANT(base_type) ==
	          TYPE_MAIN_VARIANT (unsigned_char_type_node))
	name_ptr = "UBYTE";
      else if (type == duration_timing_type_node
	       || TYPE_MAIN_VARIANT (base_type) ==
	          TYPE_MAIN_VARIANT (duration_timing_type_node))
	name_ptr = "DURATION";
      else if (type == abs_timing_type_node
	       || TYPE_MAIN_VARIANT (base_type) ==
	          TYPE_MAIN_VARIANT (abs_timing_type_node))
	name_ptr = "TIME";
      else
	name_ptr = "UINT";
    }
  else
    {
      if (base_type == chill_integer_type_node
	  || TYPE_MAIN_VARIANT (base_type) ==
	     TYPE_MAIN_VARIANT (chill_integer_type_node))
	name_ptr = "INT";
      else if (base_type == long_integer_type_node
	       || TYPE_MAIN_VARIANT (base_type) ==
	          TYPE_MAIN_VARIANT (long_integer_type_node))
	name_ptr = "LONG";
      else if (type == signed_char_type_node
	       || TYPE_MAIN_VARIANT (base_type) ==
	          TYPE_MAIN_VARIANT (signed_char_type_node))
	name_ptr = "BYTE";
      else
	name_ptr = "INT";
    }
  
  APPEND (result, name_ptr);
  
  /* see if we have a range */
  if (TREE_TYPE (type) != NULL)
    {
      mode_string = decode_constant (TYPE_MIN_VALUE (type));
      APPEND (result, mode_string->str);
      FREE (mode_string);
      APPEND (result, ":");
      mode_string = decode_constant (TYPE_MAX_VALUE (type));
      APPEND (result, mode_string->str);
      FREE (mode_string);
    }

  return result;
}

static MYSTRING *
print_struct (type, forbid)
     tree type;
     tree forbid;
{
  MYSTRING	*result = newstring ("");
  MYSTRING	*mode_string;
  tree	fields;

  if (chill_varying_type_p (type))
    {
      mode_string = grant_array_type (type);
      APPEND (result, mode_string->str);
      FREE (mode_string);
    }
  else
    {
      tree else_identifier = get_identifier ("else");
      fields = TYPE_FIELDS (type);
      
      APPEND (result, "STRUCT (");
      while (fields != NULL_TREE)
	{
	  if (TREE_CODE (TREE_TYPE (fields)) == UNION_TYPE)
	    {
	      tree variants;
	      /* Format a tagged variant record type.  */
	      APPEND (result, " CASE ");
	      if (TYPE_TAGFIELDS (TREE_TYPE (fields)) != NULL_TREE)
		{
		  tree tag_list = TYPE_TAGFIELDS (TREE_TYPE (fields));
		  for (;;)
		    {
		      tree tag_name = DECL_NAME (TREE_VALUE (tag_list));
		      APPEND (result, IDENTIFIER_POINTER (tag_name));
		      tag_list = TREE_CHAIN (tag_list);
		      if (tag_list == NULL_TREE)
			break;
		      APPEND (result, ", ");
		    }
		}
	      APPEND (result, " OF\n");
	      variants = TYPE_FIELDS (TREE_TYPE (fields));
	      
	      /* Each variant is a FIELD_DECL whose type is an anonymous
		 struct within the anonymous union.  */
	      while (variants != NULL_TREE)
		{
		  tree tag_list = TYPE_TAG_VALUES (TREE_TYPE (variants));
		  tree struct_elts = TYPE_FIELDS (TREE_TYPE (variants));
		  
		  while (tag_list != NULL_TREE)
		    {
		      tree tag_values = TREE_VALUE (tag_list);
		      APPEND (result, "   (");
		      while (tag_values != NULL_TREE)
			{
			  mode_string = get_tag_value (TREE_VALUE (tag_values));
			  APPEND (result, mode_string->str);
			  FREE (mode_string);
			  if (TREE_CHAIN (tag_values) != NULL_TREE)
			    {
			      APPEND (result, ",\n    ");
			      tag_values = TREE_CHAIN (tag_values);
			    }
			  else break;
			}
		      APPEND (result, ")");
		      tag_list = TREE_CHAIN (tag_list);
		      if (tag_list)
			APPEND (result, ",");
		      else
			break;
		    }
		  APPEND (result, " : ");
		  
		  while (struct_elts != NULL_TREE)
		    {
		      mode_string = decode_decl (struct_elts, forbid);
		      APPEND (result, mode_string->str);
		      FREE (mode_string);
		      
		      if (TREE_CHAIN (struct_elts) != NULL_TREE)
			APPEND (result, ",\n     ");
		      struct_elts = TREE_CHAIN (struct_elts);
		    }
		  
		  variants = TREE_CHAIN (variants);
		  if (variants != NULL_TREE
		      && TREE_CHAIN (variants) == NULL_TREE
		      && DECL_NAME (variants) == else_identifier)
		    {
		      tree else_elts = TYPE_FIELDS (TREE_TYPE (variants));
		      APPEND (result, "\n   ELSE ");
		      while (else_elts != NULL_TREE)
			{
			  mode_string = decode_decl (else_elts, forbid);
			  APPEND (result, mode_string->str);
			  FREE (mode_string);
			  if (TREE_CHAIN (else_elts) != NULL_TREE)
			    APPEND (result, ",\n     ");
			  else_elts = TREE_CHAIN (else_elts);
			}
		      break;
		    }
		  if (variants != NULL_TREE)
		    APPEND (result, ",\n");
		}
	      
	      APPEND (result, "\n   ESAC");
	    }
	  else
	    {
	      mode_string = decode_decl (fields, forbid);
	      APPEND (result, mode_string->str);
	      FREE (mode_string);
	    }
	  
	  fields = TREE_CHAIN (fields);
	  if (fields != NULL_TREE)
	    APPEND (result, ",\n    ");
	}
      APPEND (result, ")");
    }
  return result;
}

static MYSTRING *
print_proc_tail (type, args, forbid)
     tree type;
     tree args;
     tree forbid;
{
  MYSTRING	*result = newstring ("");
  MYSTRING	*mode_string;
  int count = 0;

  /* do the argument modes */
  for ( ; args != NULL_TREE; 
       args = TREE_CHAIN (args), count++)
    {
      char buf[20];
      tree argmode = TREE_VALUE (args);
      tree attribute = TREE_PURPOSE (args);

      if (argmode == void_type_node)
	continue;

      if (count)
	APPEND (result, ",\n       ");
      sprintf(buf, "arg%d ", count);
      APPEND (result, buf);

      if (attribute == ridpointers[(int) RID_LOC])
	argmode = TREE_TYPE (argmode);
      mode_string = get_type (argmode, forbid);
      APPEND (result, mode_string->str);
      FREE (mode_string);

      if (attribute != NULL_TREE)
	{
	  sprintf (buf, " %s", IDENTIFIER_POINTER (attribute));
	  APPEND (result, buf);
	}
    }
  APPEND (result, ")");
  
  /* return type */
  {
    tree retn_type = TREE_TYPE (type);

    if (retn_type != NULL_TREE
	&& TREE_CODE (retn_type) != VOID_TYPE)
      {
	mode_string = get_type (retn_type, forbid);
	APPEND (result, "\n  RETURNS (");
	APPEND (result, mode_string->str);
	FREE (mode_string);
	APPEND (result, ")");
      }
  }
  return result;
}


static MYSTRING *
print_proc_exceptions (ex)
     tree ex;
{
  MYSTRING	*result = newstring ("");

  if (ex != NULL_TREE)
    {
      APPEND (result, "\n  EXCEPTIONS (");
      for ( ; ex != NULL_TREE; ex = TREE_CHAIN (ex))
	{
	  APPEND (result, IDENTIFIER_POINTER (TREE_VALUE (ex)));
	  if (TREE_CHAIN (ex) != NULL_TREE)
	    APPEND (result, ",\n    ");
	}
      APPEND (result, ")");
    }
  return result;
}

/* output a mode (or type). */

static MYSTRING *
decode_mode (type, forbid)
    tree type;
    tree forbid;
{
  MYSTRING	*result = newstring ("");
  MYSTRING	*mode_string;

  switch ((enum chill_tree_code)TREE_CODE (type))
    {
    case TYPE_DECL:
      if (DECL_NAME (type))
	{
	  APPEND (result, IDENTIFIER_POINTER (DECL_NAME (type)));
	  return result;
	}
      type = TREE_TYPE (type);
      break;

    case IDENTIFIER_NODE:
      APPEND (result, IDENTIFIER_POINTER (type));
      return result;

    case LANG_TYPE:
      /* LANG_TYPE are only used until satisfy is done,
	 as place-holders for 'READ T', and NEWMODE modes. */
      if (TYPE_READONLY (type))
	APPEND (result, "READ ");

      mode_string = get_type (TREE_TYPE (type), forbid);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      break;

    case ARRAY_TYPE:
      mode_string = grant_array_type (type);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      break;

    case BOOLEAN_TYPE:
      APPEND (result, "BOOL");
      break;

    case CHAR_TYPE:
      APPEND (result, "CHAR");
      break;

    case ENUMERAL_TYPE:
      mode_string = print_enumeral (type); 
      APPEND (result, mode_string->str);
      FREE (mode_string);
      break;
	
    case FUNCTION_TYPE:
      {
	tree args = TYPE_ARG_TYPES (type);

	APPEND (result, "PROC (");

	mode_string = print_proc_tail (type, args, forbid);
	APPEND (result, mode_string->str);
	FREE (mode_string);
      }
      break;

    case INTEGER_TYPE:
      mode_string = print_integer_type (type, forbid);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      break;
	
    case RECORD_TYPE:
      if (CH_IS_INSTANCE_MODE (type))
	{
	  APPEND (result, "INSTANCE");
	  return result;
	}
      else if (CH_IS_BUFFER_MODE (type) || CH_IS_EVENT_MODE (type))
	{ tree bufsize = max_queue_size (type);
	  APPEND (result, CH_IS_BUFFER_MODE (type) ? "BUFFER " : "EVENT ");
	  if (bufsize != NULL_TREE)
	    {
	      APPEND (result, "(");
	      mode_string = decode_constant (bufsize);
	      APPEND (result, mode_string->str);
	      APPEND (result, ") ");
	      FREE (mode_string);
	    }
	  if (CH_IS_BUFFER_MODE (type))
	    {
	      mode_string = decode_mode (buffer_element_mode (type), forbid);
	      APPEND (result, mode_string->str);
	      FREE (mode_string);
	    }
	  break;
	}      
      mode_string = print_struct (type, forbid);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      break;

    case POINTER_TYPE:
      if (TREE_CODE (TREE_TYPE (type)) == VOID_TYPE)
	APPEND (result, "PTR");
      else
	{
	  if (TREE_CODE (TREE_TYPE (type)) == FUNCTION_TYPE)
	    {
	      mode_string = get_type (TREE_TYPE (type), forbid);
	      APPEND (result, mode_string->str);
	      FREE (mode_string);
	    }
	  else
	    {
	      APPEND (result, "REF ");
	      mode_string = get_type (TREE_TYPE (type), forbid);
	      APPEND (result, mode_string->str);
	      FREE (mode_string);
	    }
	}
      break;

    case REAL_TYPE:
      if (TREE_INT_CST_LOW (TYPE_SIZE (type)) == 32)
	APPEND (result, "REAL");
      else
	APPEND (result, "LONG_REAL");
      break;

    case SET_TYPE:
      if (CH_BOOLS_TYPE_P (type))
	mode_string = grant_array_type (type);
      else
	{
	  APPEND (result, "POWERSET ");
	  mode_string = get_type (TYPE_DOMAIN (type), forbid);
	}
      APPEND (result, mode_string->str);
      FREE (mode_string);
      break;
	
    case REFERENCE_TYPE:
      mode_string = get_type (TREE_TYPE (type), forbid);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      break;
      
    default:
      APPEND (result, "/* ---- not implemented ---- */");
      break;
    }

  return (result);
}

static MYSTRING *
get_type (type, forbid)
    tree	type;
    tree	forbid;
{
  if (type == NULL_TREE || TREE_CODE (type) == ERROR_MARK)
    return newstring ("");

  return (decode_mode (type, forbid));
}

static int
is_forbidden (str, forbid)
    tree	str;
    tree	forbid;
{
  if (forbid == NULL_TREE)
    return (0);
  
  if (TREE_CODE (forbid) == INTEGER_CST)
    return (1);
  
  while (forbid != NULL_TREE)
    {
      if (TREE_VALUE (forbid) == str)
	return (1);
      forbid = TREE_CHAIN (forbid);
    }
  /* nothing found */
  return (0);
}

static MYSTRING *
decode_constant (init)
     tree	init;
{
  MYSTRING *result = newstring ("");
  MYSTRING *tmp_string;
  tree	    type = TREE_TYPE (init);
  tree	val = init;
  char *op;
  char	wrk[256];
  MYSTRING *mode_string;
    
  switch ((enum chill_tree_code)TREE_CODE (val))
    {
    case CALL_EXPR:
      tmp_string = decode_constant (TREE_OPERAND (val, 0));
      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      val = TREE_OPERAND (val, 1);  /* argument list */
      if (val != NULL_TREE && TREE_CODE (val) != TREE_LIST)
	{
	  APPEND (result, " ");
	  tmp_string = decode_constant (val);
	  APPEND (result, tmp_string->str);
	  FREE (tmp_string);
	}
      else
	{
	  APPEND (result, " (");
	  if (val != NULL_TREE)
	    {
	      for (;;)
		{
		  tmp_string = decode_constant (TREE_VALUE (val));
		  APPEND (result, tmp_string->str);
		  FREE (tmp_string);
		  val = TREE_CHAIN (val);
		  if (val == NULL_TREE)
		    break;
		  APPEND (result, ", ");
		}
	    }
	  APPEND (result, ")");
	}
      return result;

    case NOP_EXPR:
      /* Generate an "expression conversion" expression (a cast). */
      tmp_string = decode_mode (type, NULL_TREE);

      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      APPEND (result, "(");
      val = TREE_OPERAND (val, 0);
      type = TREE_TYPE (val);

      /* If the coercee is a tuple, make sure it is prefixed by its mode. */
      if (TREE_CODE (val) == CONSTRUCTOR
	&& !CH_BOOLS_TYPE_P (type) && !chill_varying_type_p (type))
	{
	  tmp_string = decode_mode (type, NULL_TREE);
	  APPEND (result, tmp_string->str);
	  FREE (tmp_string);
	  APPEND (result, " ");
	}

      tmp_string = decode_constant (val);
      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      APPEND (result, ")");
      return result;

    case IDENTIFIER_NODE:
      APPEND (result, IDENTIFIER_POINTER (val));
      return result;

    case PAREN_EXPR:
      APPEND (result, "(");
      tmp_string = decode_constant (TREE_OPERAND (val, 0));
      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      APPEND (result, ")");
      return result;

    case UNDEFINED_EXPR:
      APPEND (result, "*");
      return result;

    case PLUS_EXPR:        op = "+";       goto binary;
    case MINUS_EXPR:       op = "-";       goto binary;
    case MULT_EXPR:        op = "*";       goto binary;
    case TRUNC_DIV_EXPR:   op = "/";       goto binary;
    case FLOOR_MOD_EXPR:   op = " MOD ";   goto binary;
    case TRUNC_MOD_EXPR:   op = " REM ";   goto binary;
    case CONCAT_EXPR:      op = "//";      goto binary;
    case BIT_IOR_EXPR:     op = " OR ";    goto binary;
    case BIT_XOR_EXPR:     op = " XOR ";   goto binary;
    case TRUTH_ORIF_EXPR:  op = " ORIF ";  goto binary;
    case BIT_AND_EXPR:     op = " AND ";   goto binary;
    case TRUTH_ANDIF_EXPR: op = " ANDIF "; goto binary;
    case GT_EXPR:          op = ">";       goto binary;
    case GE_EXPR:          op = ">=";      goto binary;
    case SET_IN_EXPR:      op = " IN ";    goto binary;
    case LT_EXPR:          op = "<";       goto binary;
    case LE_EXPR:          op = "<=";      goto binary;
    case EQ_EXPR:          op = "=";       goto binary;
    case NE_EXPR:          op = "/=";      goto binary;
    case RANGE_EXPR:
      if (TREE_OPERAND (val, 0) == NULL_TREE)
	{
	  APPEND (result, "*");
	  return result;
	}
      op = ":";       goto binary;
    binary:
      tmp_string = decode_constant (TREE_OPERAND (val, 0));
      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      APPEND (result, op);
      tmp_string = decode_constant (TREE_OPERAND (val, 1));
      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      return result;

    case REPLICATE_EXPR:
      APPEND (result, "(");
      tmp_string = decode_constant (TREE_OPERAND (val, 0));
      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      APPEND (result, ")");
      tmp_string = decode_constant (TREE_OPERAND (val, 1));
      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      return result;

    case NEGATE_EXPR:     op = "-";     goto unary;
    case BIT_NOT_EXPR:    op = " NOT "; goto unary;
    case ADDR_EXPR:       op = "->"; goto unary;
    unary:
      APPEND (result, op);
      tmp_string = decode_constant (TREE_OPERAND (val, 0));
      APPEND (result, tmp_string->str);
      FREE (tmp_string);
      return result;

    case INTEGER_CST:
      APPEND (result, display_int_cst (val));
      return result;

#ifndef REAL_IS_NOT_DOUBLE
    case REAL_CST:
      sprintf (wrk, "%.20g", TREE_REAL_CST (val));
      APPEND (result, wrk);
      return result;
#endif

    case STRING_CST:
      {
	char *ptr = TREE_STRING_POINTER (val);
	int i = TREE_STRING_LENGTH (val);
	APPEND (result, "\"");
	while (--i >= 0)
	  {
	    char buf[10];
	    char c = *ptr++;
	    if (c == '^')
	      APPEND (result, "^^");
	    else if (c == '"')
	      APPEND (result, "\"\"");
	    else if (c == '\n')
	      APPEND (result, "^J");
	    else
	      {
		buf[0] = c;
		buf[1] = 0;
		APPEND (result, buf);
	      }
	  }
	APPEND (result, "\"");
	return result;
      }

    case CONSTRUCTOR:
      val = TREE_OPERAND (val, 1);
      if (type != NULL && TREE_CODE (type) == SET_TYPE
	  && CH_BOOLS_TYPE_P (type))
	{
	  /* It's a bitstring. */
	  tree domain = TYPE_DOMAIN (type);
	  tree domain_max = TYPE_MAX_VALUE (domain);
	  char *buf;
	  register char *ptr;
	  int len;
	  if (TREE_CODE (domain_max) != INTEGER_CST
	      || (val && TREE_CODE (val) != TREE_LIST))
	    goto fail;

	  len = TREE_INT_CST_LOW (domain_max) + 1;
	  if (TREE_CODE (init) != CONSTRUCTOR)
	    goto fail;
	  buf = (char *) alloca (len + 10);
	  ptr = buf;
	  *ptr++ = ' ';	  
	  *ptr++ = 'B';
	  *ptr++ = '\'';
	  if (get_set_constructor_bits (init, ptr, len))
	    goto fail;
	  for (; --len >= 0; ptr++)
	    *ptr += '0';
	  *ptr++ = '\'';
	  *ptr = '\0';
	  APPEND (result, buf);
	  return result;
	}
      else
	{ /* It's some kind of tuple */
	  if (type != NULL_TREE)
	    {
	      mode_string = get_type (type, NULL_TREE);
	      APPEND (result, mode_string->str);
	      FREE (mode_string);
	      APPEND (result, " ");
	    }
	  if (val == NULL_TREE
	      || TREE_CODE (val) == ERROR_MARK)
	    APPEND (result, "[ ]");
	  else if (TREE_CODE (val) != TREE_LIST)
	    goto fail;
	  else
	    {
	      APPEND (result, "[");
	      for ( ; ; )
		{
		  tree lo_val = TREE_PURPOSE (val);
		  tree hi_val = TREE_VALUE (val);
		  MYSTRING *val_string;
		  if (TUPLE_NAMED_FIELD (val))
		    APPEND(result, ".");
		  if (lo_val != NULL_TREE)
		    {
		      val_string = decode_constant (lo_val);
		      APPEND (result, val_string->str);
		      FREE (val_string);
		      APPEND (result, ":");
		    }
		  val_string = decode_constant (hi_val);
		  APPEND (result, val_string->str);
		  FREE (val_string);
		  val = TREE_CHAIN (val);
		  if (val == NULL_TREE)
		    break;
		  APPEND (result, ", ");
		}
	      APPEND (result, "]");
	    }
	}
      return result;
    fail:
      error ("decode_constant: mode and value mismatch");
      break;
    default:
      error ("decode_constant: cannot decode this mode");
      break;
    }
  return result;
}

/* Assuming DECL is an ALIAS_DECL, return its prefix rename clause. */

static MYSTRING *
decode_prefix_rename (decl)
    tree decl;
{
  MYSTRING *result = newstring ("");
  if (DECL_OLD_PREFIX (decl) || DECL_NEW_PREFIX (decl))
    {
      APPEND (result, "(");
      if (DECL_OLD_PREFIX (decl))
	APPEND (result, IDENTIFIER_POINTER (DECL_OLD_PREFIX (decl)));
      APPEND (result, "->");
      if (DECL_NEW_PREFIX (decl))
 	APPEND (result, IDENTIFIER_POINTER (DECL_NEW_PREFIX (decl)));
      APPEND (result, ")!");
    }
  if (DECL_POSTFIX_ALL (decl))
    APPEND (result, "ALL");
  else
    APPEND (result, IDENTIFIER_POINTER  (DECL_POSTFIX (decl)));
  return result;
}

static MYSTRING *
decode_decl (decl, forbid)
    tree decl;
    tree forbid;
{
  MYSTRING *result = newstring ("");
  MYSTRING *mode_string;
  tree      type;
  
  switch ((enum chill_tree_code)TREE_CODE (decl))
    {
    case VAR_DECL:
    case BASED_DECL:
      APPEND (result, "DCL ");
      APPEND (result, IDENTIFIER_POINTER (DECL_NAME (decl)));
      APPEND (result, " ");
      mode_string = get_type (TREE_TYPE (decl), forbid);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      if ((enum chill_tree_code)TREE_CODE (decl) == BASED_DECL)
        {
          APPEND (result, " BASED (");
          APPEND (result, IDENTIFIER_POINTER (DECL_ABSTRACT_ORIGIN (decl)));
          APPEND (result, ")");
        }
      break;

    case TYPE_DECL:
      if (CH_DECL_SIGNAL (decl))
	{
	  /* this is really a signal */
	  tree fields = TYPE_FIELDS (TREE_TYPE (decl));
	  tree signame = DECL_NAME (decl);
	  tree sigdest;
	  
	  APPEND (result, "SIGNAL ");
	  APPEND (result, IDENTIFIER_POINTER (signame));
	  if (IDENTIFIER_SIGNAL_DATA (signame))
	    {
	      APPEND (result, " = (");
	      for ( ; fields != NULL_TREE;
		   fields = TREE_CHAIN (fields))
		{
		  MYSTRING *mode_string;
		  
		  mode_string = get_type (TREE_TYPE (fields),
					  NULL_TREE);
		  APPEND (result, mode_string->str);
		  FREE (mode_string);
		  if (TREE_CHAIN (fields) != NULL_TREE)
		    APPEND (result, ", ");
		}
	      APPEND (result, ")");
	    }
	  sigdest = IDENTIFIER_SIGNAL_DEST (signame);
	  if (sigdest != NULL_TREE)
	    {
	      APPEND (result, " TO ");
	      APPEND (result, IDENTIFIER_POINTER (DECL_NAME (sigdest)));
	    }
	}
      else
	{
	  /* avoid defining a mode as itself */
	  if (CH_NOVELTY_FLAG (decl))
	    APPEND (result, "NEWMODE ");
	  else
	    APPEND (result, "SYNMODE ");
	  APPEND (result, IDENTIFIER_POINTER (DECL_NAME (decl)));
	  APPEND (result, " = ");
	  mode_string = decode_mode (TREE_TYPE (decl), forbid);
	  APPEND (result, mode_string->str);
	  FREE (mode_string);
	}
      break;
      
    case FUNCTION_DECL:
      {
	tree	args, ex;
	
	type = TREE_TYPE (decl);
	args = TYPE_ARG_TYPES (type);
	ex = TYPE_RAISES_EXCEPTIONS (type);
	
	APPEND (result, IDENTIFIER_POINTER (DECL_NAME (decl)));
	
	if (CH_DECL_PROCESS (decl))
	  APPEND (result, ": PROCESS (");
	else
	  APPEND (result, ": PROC (");

	args = TYPE_ARG_TYPES (type);
	
	mode_string = print_proc_tail (type, args, forbid);
	APPEND (result, mode_string->str);
	FREE (mode_string);
	
	mode_string = print_proc_exceptions (ex);
	APPEND (result, mode_string->str);
	FREE (mode_string);
	
	/* generality */
	if (CH_DECL_GENERAL (decl))
	  APPEND (result, " GENERAL");
	if (CH_DECL_SIMPLE (decl))
	  APPEND (result, " SIMPLE");
	if (DECL_INLINE (decl))
	  APPEND (result, " INLINE");
	if (CH_DECL_RECURSIVE (decl))
	  APPEND (result, " RECURSIVE");
	APPEND (result, " END");
      }
      break;
      
    case FIELD_DECL:
      APPEND (result, IDENTIFIER_POINTER (DECL_NAME (decl)));
      APPEND (result, " ");
      mode_string = get_type (TREE_TYPE (decl), NULL_TREE);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      if (is_forbidden (DECL_NAME (decl), forbid))
	APPEND (result, " FORBID");
      break;
      
    case CONST_DECL:
      if (DECL_INITIAL (decl) == NULL_TREE 
	  || TREE_CODE (DECL_INITIAL (decl)) == ERROR_MARK)
	break;
      APPEND (result, "SYN ");
      APPEND (result, IDENTIFIER_POINTER (DECL_NAME (decl)));
      APPEND (result, " ");
      mode_string = get_type (TREE_TYPE (decl), forbid);
      APPEND (result, mode_string->str);
      FREE (mode_string);
      APPEND (result, " = ");
      mode_string = decode_constant (DECL_INITIAL (decl));
      APPEND (result, mode_string->str);
      FREE (mode_string);
      break;
      
    case ALIAS_DECL:
      /* If CH_DECL_GRANTED, decl was granted into this scope, and
	 so wasn't in the source code. */
      if (!CH_DECL_GRANTED (decl))
	{
	  if (DECL_SEIZEFILE (decl) != use_seizefile_name
	      && DECL_SEIZEFILE (decl))
	    {
	      use_seizefile_name = DECL_SEIZEFILE (decl);
	      grant_use_seizefile (IDENTIFIER_POINTER (use_seizefile_name));
	      mark_use_seizefile_written (use_seizefile_name);
	    }
	  APPEND (result, "SEIZE ");
	  mode_string = decode_prefix_rename (decl);
	  APPEND (result, mode_string->str);
	  FREE (mode_string);
	}
      break;

    default:
      APPEND (result, "----- not implemented ------");
      break;
    }
  return (result);
}

static void
globalize_decl (decl)
    tree	decl;
{
  if (!TREE_PUBLIC (decl) && DECL_NAME (decl) &&
      (TREE_CODE (decl) == VAR_DECL || TREE_CODE (decl) == FUNCTION_DECL))
    {
      extern	FILE	*asm_out_file;
      extern	char	*first_global_object_name;
      char		*name = XSTR (XEXP (DECL_RTL (decl), 0), 0);
	
      if (!first_global_object_name)
	first_global_object_name = name + (name[0] == '*');
      ASM_GLOBALIZE_LABEL (asm_out_file, name);
    }
}


static void
grant_one_decl (decl)
    tree	decl;
{
  MYSTRING	*result;

  if (DECL_SOURCE_LINE (decl) == 0)
    return;
  result = decode_decl (decl, IDENTIFIER_FORBID (DECL_NAME (decl)));
  if (result->len)
    {
      APPEND (result, ";\n");
      APPEND (gstring, result->str);
    }
  FREE (result);
}

static int
compare_memory_file (fname, buf)
    char	*fname;
    char	*buf;
{
  FILE	*fb;
  int		c;

  /* check if we have something to write */
  if (!buf || !strlen (buf))
    return (0);
    
  if ((fb = fopen (fname, "r")) == NULL)
    return (1);
    
  while ((c = getc (fb)) != EOF)
    {
      if (c != *buf++)
	{
	  fclose (fb);
	  return (1);
	}
    }
  fclose (fb);
  return (*buf ? 1 : 0);
}

void
write_grant_file ()
{
  FILE	*fb;

  /* We only write out the grant file if it has changed,
     to avoid changing its time-stamp and triggering an
     unnecessary 'make' action.  Return if no change. */
  if (gstring == NULL || !spec_module_generated ||
      !compare_memory_file (grant_file_name, gstring->str))
    return;

  fb = fopen (grant_file_name, "w");
  if (fb == NULL)
      pfatal_with_name (grant_file_name);
    
  /* write file. Due to problems with record sizes on VAX/VMS
     write string to '\n' */
#ifdef VMS
  /* do it this way for VMS, cause of problems with
     record sizes */
  p = gstring->str;
  while (*p)
    {
      extern char* strchr ();
      p1 = strchr (p, '\n');
      c = *++p1;
      *p1 = '\0';
      fprintf (fb, "%s", p);
      *p1 = c;
      p = p1;
    }
#else
  /* faster way to write */
  if (write (fileno (fb), gstring->str, gstring->len) < 0)
    {
      int save_errno = errno;
      unlink (grant_file_name);
      errno = save_errno;
      pfatal_with_name (grant_file_name);
    }
#endif
  fclose (fb);
}


/* handle grant statement */

void
set_default_grant_file ()
{
#undef strrchr
    extern	char *strrchr ();
    char	*p, *tmp, *fname;

    if (dump_base_name)
      fname = dump_base_name; /* Probably invoked via gcc */
    else
      { /* Probably invoked directly (not via gcc) */
	fname = asm_file_name;
	if (!fname)
	  fname = main_input_filename ? main_input_filename : input_filename;
	if (!fname)
	  return;
      }

    p = strrchr (fname, '.');
    if (!p)
    {
	tmp = (char *) alloca (strlen (fname) + 10);
	strcpy (tmp, fname);
    }
    else
    {
	int	i = p - fname;
	
	tmp = (char *) alloca (i + 10);
	strncpy (tmp, fname, i);
	tmp[i] = '\0';
    }
    strcat (tmp, ".grt");
    default_grant_file = build_string (strlen (tmp), tmp);

    grant_file_name = TREE_STRING_POINTER (default_grant_file);

    if (gstring == NULL)
      gstring = newstring ("");
}

/* Make DECL visible under the name NAME in the (fake) outermost scope. */

void
push_granted (name, decl)
     tree name, decl;
{
#if 0
  IDENTIFIER_GRANTED_VALUE (name) = decl;
  granted_decls = tree_cons (name, decl, granted_decls);
#endif
}

void
chill_grant (old_prefix, new_prefix, postfix, forbid)
     tree old_prefix;
     tree new_prefix;
     tree postfix;
     tree forbid;
{
  if (pass == 1)
    {
#if 0
      tree old_name = old_prefix == NULL_TREE ? postfix
	: get_identifier3 (IDENTIFIER_POINTER (old_prefix),
			   "!", IDENTIFIER_POINTER (postfix));
      tree new_name = new_prefix == NULL_TREE ? postfix
	: get_identifier3 (IDENTIFIER_POINTER (new_prefix),
			   "!", IDENTIFIER_POINTER (postfix));
#endif
      tree alias = build_alias_decl (old_prefix, new_prefix, postfix);
      CH_DECL_GRANTED (alias) = 1;
      DECL_SEIZEFILE (alias) = current_seizefile_name;
      TREE_CHAIN (alias) = current_module->granted_decls;
      current_module->granted_decls = alias;

      if (forbid)
	warning ("FORBID is not yet implemented");  /* FIXME */
    }
}

/* Write a SPEC MODULE using the declarations in the list DECLS. */

static char *header_template =
"--\n-- WARNING: this file was generated by\n\
-- GNUCHILL version %s\n-- based on gcc version %s\n--\n";

void
write_spec_module (decls, granted_decls)
     tree decls;
     tree granted_decls;
{
  tree	 vars;
  char   *hdr;

  if (granted_decls == NULL_TREE)
    return;
  
  use_seizefile_name = NULL_TREE;

  hdr = (char*) alloca (strlen (gnuchill_version)
			+ strlen (version_string)
			+ strlen (header_template) + 1);
  sprintf (hdr, header_template, gnuchill_version, version_string);
  APPEND (gstring, hdr);
  
  APPEND (gstring, IDENTIFIER_POINTER (chill_module_name));
  APPEND (gstring, ": SPEC MODULE\n");
  
  for (vars = decls; vars != NULL_TREE; vars = TREE_CHAIN (vars))
    {
      if (DECL_SOURCE_LINE (vars))
        {
          if (DECL_NAME (vars))
            {
              if (TREE_CODE (vars) != CONST_DECL || !CH_DECL_ENUM (vars))
                grant_one_decl (vars);
            }
          else if (DECL_POSTFIX_ALL (vars))
            {
              if (DECL_SEIZEFILE (vars) != use_seizefile_name
                  && DECL_SEIZEFILE (vars))
                {
                  use_seizefile_name = DECL_SEIZEFILE (vars);
                  grant_use_seizefile (IDENTIFIER_POINTER (use_seizefile_name));
                  mark_use_seizefile_written (use_seizefile_name);
                }
              APPEND (gstring, "SEIZE ALL;\n");
            }
        }
    }

  for (vars = granted_decls; vars != NULL_TREE; vars = TREE_CHAIN (vars))
    if (DECL_SOURCE_LINE (vars))
      {
	MYSTRING *mode_string = decode_prefix_rename (vars);
	APPEND (gstring, "GRANT ");
	APPEND (gstring, mode_string->str);
	FREE (mode_string);
	APPEND (gstring, ";\n");
      }

  APPEND (gstring, "END;\n");
  spec_module_generated = 1;
}

/*
 * This function is called from the parser to preface the first
 * module-level action with the beginning of the function which
 * will enclose that action.
 */
void
chill_start_module_code ()
{
    char *init_name = (char *)alloca (IDENTIFIER_LENGTH (chill_module_name) + 10);

    /* FIXME: need to handle module nested inside a module */
    if (!global_bindings_p() || module_function_name)
      return;
    
    /* Add an underscore to distinguish the real module-body code
       (inside the following function) from the init code
       (inside a function not yet created by finish_module_code). */
    strcpy (init_name, IDENTIFIER_POINTER (chill_module_name));
    strcat (init_name, "_");
    module_function_name = get_identifier (init_name);

    start_chill_function (module_function_name,
			  void_type_node,
			  NULL_TREE, NULL_TREE, NULL_TREE);
    DECL_SOURCE_LINE (current_function_decl) = 0;
   
    chill_at_module_level = 1;
}    

/* this function finishes the function for code at module level
   and builds an init_list entry for the module's module-level
   code, if there was any. */
void
chill_finish_module_code (module_id)
     tree module_id;
{
  if (module_function_name == NULL_TREE)
    return;

  /* Every module has a name, generated by push_module if
     not by the user. */
  if (module_id == NULL_TREE)
    abort ();

  /* If there was module-level code in this module (not just function
     declarations), we allocate space for this module's init list entry,
     and fill in the module's function's address. */
  if (module_function_name != NULL_TREE)
    {
      extern tree initializer_type;
      char *module_id_str = IDENTIFIER_POINTER (module_id);
      char *init_entry_name = (char *)xmalloc ((unsigned)(strlen (module_id_str) + 20));
      tree  init_entry_id;
      tree  init_entry_decl;
      tree  module_function_decl = current_function_decl;
      
      finish_chill_function (0);

      sprintf (init_entry_name, "__tmp_%s_init_entry", 
	       module_id_str);
      init_entry_id = get_identifier (init_entry_name);
      
      init_entry_decl = build1 (ADDR_EXPR,
				TREE_TYPE (TYPE_FIELDS (initializer_type)),
				module_function_decl);
      TREE_CONSTANT (init_entry_decl) = 1;
      init_entry_decl
        = do_decl (init_entry_id, initializer_type, 1, 
	    build (CONSTRUCTOR, initializer_type, NULL_TREE,
	      tree_cons (NULL_TREE, init_entry_decl,
		  tree_cons (NULL_TREE, null_pointer_node,
			     NULL_TREE))),
		   0);
      DECL_SOURCE_LINE (init_entry_decl) = 0;
      if (pass == 1)
	/* tell chill_finish_compile that there's 
 	   module-level code to be processed. */
	module_init_list = integer_one_node;
      else
	module_init_list = tree_cons (module_function_decl,
				      init_entry_decl,
				      module_init_list);
    }
  
  chill_at_module_level = 0;
  module_function_name = NULL_TREE;
}

/*
 * after the dark comes, after all of the modules are at rest,
 * we tuck the compilation unit to bed...  A story in pass 1
 * and a hug-and-a-kiss goodnight in pass 2.
 */
void
chill_finish_compile ()
{
  tree global_list;
  tasking_setup ();
  build_enum_tables ();
  
  /* We only need an initializer function for the source file if
     a) there's module-level code to be called, or
     b) tasking-related stuff to be initialized. */
  if (module_init_list != NULL_TREE || tasking_list != NULL_TREE)
    {
      extern tree initializer_type;
      static tree chill_init_name;

      /* declare the global initializer list */
      global_list = do_decl (get_identifier ("_ch_init_list"),
			     build_chill_pointer_type (initializer_type), 1, 
			     NULL_TREE, 1);

      /* Now, we're building the function which is the *real*
	 constructor - if there's any module-level code in this
	 source file, the compiler puts the file's initializer entry
	 onto the global initializer list, so each module's body code
	 will eventually get called, after all of the processes have
	 been started up.  */
      
      /* This is better done in pass 2 (when first_global_object_name
	 may have been set), but that is too late.
	 Perhaps rewrite this so nothing is done in pass 1. */
      if (pass == 1)
	{
	  extern char *first_global_object_name;
	  /* If we don't do this spoof, we get the name of the first
	     tasking_code variable, and not the file name. */
	  char *tmp = first_global_object_name;

	  first_global_object_name = NULL;
	  chill_init_name = get_file_function_name ('I');
	  first_global_object_name = tmp;
	  /* strip off the file's extension, if any. */
	  tmp = strrchr (IDENTIFIER_POINTER (chill_init_name), '.');
	  if (tmp)
	    *tmp = '\0';
	}

      start_chill_function (chill_init_name, void_type_node, NULL_TREE,
			    NULL_TREE, NULL_TREE);
      TREE_PUBLIC (current_function_decl) = 1;
      
      /* For each module that we've compiled, that had module-level 
	 code to be called, add its entry to the global initializer
	 list. */
	 
      if (pass == 2)
	{
	  tree module_init;

	  for (module_init = module_init_list;  
	       module_init != NULL_TREE;
	       module_init = TREE_CHAIN (module_init))
	    {
	      tree init_entry      = TREE_VALUE (module_init);
	      /* tree module_function = TREE_PURPOSE (module_init); */

	      /* assign module_entry.next := _ch_init_list; */
	      expand_expr_stmt (
		build_chill_modify_expr (
		  build_component_ref (init_entry,
		    get_identifier ("__INIT_NEXT")),
		      global_list));

	      /* assign _ch_init_list := &module_entry; */
	      expand_expr_stmt (
		build_chill_modify_expr (global_list,
		  build1 (ADDR_EXPR, ptr_type_node, init_entry)));
	    }
	}

      tasking_registry ();

      finish_chill_function (0);

      if (pass == 2)
	{
	  tree decl = lookup_name (chill_init_name);
	  if (decl != NULL_TREE)
	    {
	      assemble_constructor (IDENTIFIER_POINTER (chill_init_name));
	      globalize_decl (lookup_name (chill_init_name));
	    }
	}

      /* ready now to link decls onto this list in pass 2. */
      module_init_list = NULL_TREE;
      tasking_list = NULL_TREE;
    }
}
