/* Implement I/O-related actions for CHILL.
   Copyright (C) 1992, 93, 1994 Free Software Foundation, Inc.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <stdio.h>
#include <limits.h>
#include "config.h"
#include "tree.h"
#include "ch-tree.h"
#include "rtl.h"
#include "lex.h"
#include "flags.h"
#include "input.h"
#include "assert.h"

/* set non-zero if input text is forced to lowercase */
extern int ignore_case;

/* set non-zero if special words are to be entered in uppercase */
extern int special_UC;

extern void error   PROTO((char *, ...));
extern void sorry   PROTO((char *, ...));
extern void warning PROTO((char *, ...));

extern tree build_chill_compound_expr PROTO((tree));

/* we have to distinguish between io-list-type for WRITETEXT
   and for READTEXT. WRITETEXT does not process ranges and
   READTEXT must get pointers to the variables.
*/
/* variable to hold the type of io_list for WRITETEXT */
static tree write_io_list_type = NULL_TREE;
/* variable to hold the type of io_list for READTEXT */
static tree read_io_list_type = NULL_TREE;

/* the type for the enum tables */
static tree enum_table_type = NULL_TREE;

/* structure to save enums for later use in compilation */
typedef struct save_enum_names
{
    struct save_enum_names	*forward;
    tree			name;
    tree			decl;
} SAVE_ENUM_NAMES;

static SAVE_ENUM_NAMES *used_enum_names = (SAVE_ENUM_NAMES *)0;

typedef struct save_enum_values
{
    long			val;
    struct save_enum_names	*name;
} SAVE_ENUM_VALUES;

typedef struct save_enums
{
    struct save_enums		*forward;
    tree			context;
    tree			type;
    tree			ptrdecl;
    long			num_vals;
    struct save_enum_values	*vals;
} SAVE_ENUMS;

static SAVE_ENUMS	*used_enums = (SAVE_ENUMS *)0;


/* Function collects all enums are necessary to collect, makes a copy of
   the value and returns a VAR_DECL external to current function describing
   the pointer to a name table, which will be generated at the end of
   compilation
*/

static tree add_enum_to_list (type, context)
    tree  type;
    tree  context;
{
    tree		tmp;
    SAVE_ENUMS		*wrk = used_enums;
    SAVE_ENUM_VALUES	*vals;
    SAVE_ENUM_NAMES	*names;
    
    while (wrk != (SAVE_ENUMS *)0)
    {
	/* search for this enum already in use */
	if (wrk->context == context && wrk->type == type)
	{
	    /* yes, found. look if the ptrdecl is valid in this scope */
	    char  *name = IDENTIFIER_POINTER (DECL_NAME (wrk->ptrdecl));
	    tree   var  = get_identifier (name);
	    tree   decl = lookup_name (var);
	    
	    if (decl == NULL_TREE)
	    {
		/* no, not valid in this context, declare it */
		decl = decl_temp1 (var, build_pointer_type (TREE_TYPE (enum_table_type)),
				   0, NULL_TREE, 1, 0);
	    }
	    return decl;
	}
	
	/* next one */
	wrk = wrk->forward;
    }
    
    /* not yet found -- generate an entry */
    wrk = (SAVE_ENUMS *)xmalloc (sizeof (SAVE_ENUMS));
    wrk->forward = used_enums;
    used_enums = wrk;
    
    /* generate the pointer decl */
    wrk->ptrdecl = get_unique_identifier ("ENUMTABPTR");
    wrk->ptrdecl = decl_temp1 (wrk->ptrdecl, build_pointer_type (TREE_TYPE (enum_table_type)),
			       0, NULL_TREE, 1, 0);

    /* save information for later use */
    wrk->context = context;
    wrk->type = type;

    /* insert the names and values */
    tmp = TYPE_FIELDS (type);
    wrk->num_vals = list_length (tmp);
    vals = (SAVE_ENUM_VALUES *)xmalloc (sizeof (SAVE_ENUM_VALUES) * wrk->num_vals);
    wrk->vals = vals;
    
    while (tmp != NULL_TREE)
    {
	/* search if name is already in use */
	names = used_enum_names;
	while (names != (SAVE_ENUM_NAMES *)0)
	{
	    if (names->name == TREE_PURPOSE (tmp))
		break;
	    names = names->forward;
	}
	if (names == (SAVE_ENUM_NAMES *)0)
	{
	    /* we have to insert one */
	    names = (SAVE_ENUM_NAMES *)xmalloc (sizeof (SAVE_ENUM_NAMES));
	    names->forward = used_enum_names;
	    used_enum_names = names;
	    names->decl = NULL_TREE;
	    names->name = TREE_PURPOSE (tmp);
	}
	vals->name = names;
	vals->val = TREE_INT_CST_LOW (TREE_VALUE (tmp));
	
	/* next entry in enum */
	vals++;
	tmp = TREE_CHAIN (tmp);
    }
    
    /* return the generated decl */
    return wrk->ptrdecl;
}


void
build_chill_io_list_type ()
{
    tree list = NULL_TREE;
    tree result, enum1, listbase;
    tree io_descriptor;
    tree decl1, decl2;
    tree forcharstring, forset;
#if 0
    tree forpowerset;
#endif
    tree forintrange, foruintrange, forsetrange, forcharrange;
    tree long_type, ulong_type, union_type;
    /* We temporarily reset the maximum_field_alignment to zero so the
       compiler's init data structures can be compatible with the
       run-time system, even when we're compiling with -fpack. */
    extern int maximum_field_alignment;
    int save_maximum_field_alignment = maximum_field_alignment;
  maximum_field_alignment = 0;
    
    long_type = TREE_TYPE (lookup_name (
                           get_identifier ((ignore_case || ! special_UC) ? 
					   "long" : "LONG")));
    
    ulong_type = TREE_TYPE (lookup_name (
			    get_identifier ((ignore_case || ! special_UC) ? 
					    "ulong" : "ULONG")));

    if (write_io_list_type != NULL_TREE)
	/* already done */
	goto build_read_io_list_type;

    /* build type for enum_tables */
    decl1 = build_decl (FIELD_DECL, get_identifier ("value"),
			long_type);
    DECL_INITIAL (decl1) = NULL_TREE;
    decl2 = build_decl (FIELD_DECL, get_identifier ("name"),
			build_pointer_type (char_type_node));
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    result = build_chill_struct_type (decl1);
    pushdecl (enum_table_type = build_decl (TYPE_DECL,
					    get_identifier ("__tmp_IO_enum_table_type"),
					    result));
    DECL_SOURCE_LINE (enum_table_type) = 0;
    satisfy_decl (enum_table_type, 0);

    /* first build the enum for the desriptor */
    enum1 = start_enum (NULL_TREE);
    result = build_enumerator (get_identifier ("_IO_UNUSED"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_Byte"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_UByte"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_Int"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_UInt"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_Long"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_ULong"),
			       NULL_TREE);
    list = chainon (result, list);

    result = build_enumerator (get_identifier ("_IO_Bool"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_Char"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_CharRange"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_CharStr"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_CharVarying"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_BitStr"),
			       NULL_TREE);
    list = chainon (result, list);
#if 0
    result = build_enumerator (get_identifier ("_IO_Powerset"),
			       NULL_TREE);
    list = chainon (result, list);
#endif
    result = build_enumerator (get_identifier ("_IO_IntRange"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_UIntRange"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_Set"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_SetRange"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_Real"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_LongReal"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_Pointer"),
			       NULL_TREE);
    list = chainon (result, list);
    
    result = build_enumerator (get_identifier ("_IO_Instance"),
			       NULL_TREE);
    list = chainon (result, list);

    result = finish_enum (enum1, list);
    pushdecl (io_descriptor = build_decl (TYPE_DECL,
				get_identifier ("__tmp_IO_enum"),
				  result));
    /* prevent seizing/granting of the decl */
    DECL_SOURCE_LINE (io_descriptor) = 0;
    satisfy_decl (io_descriptor, 0);

    /* build structure for character string */
    decl1 = build_decl (FIELD_DECL, 
	      get_identifier ("string"),
		build_pointer_type (char_type_node));
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, 
	      get_identifier ("string_length"),
		ulong_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (forcharstring = build_decl (TYPE_DECL,
      get_identifier ("__tmp_IO_forcharstring"), result));
    DECL_SOURCE_LINE (forcharstring) = 0;
    satisfy_decl (forcharstring, 0);

    /* build type for set mode */
    decl1 = build_decl (FIELD_DECL, get_identifier ("value"),
			long_type);
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("name_table"),
			build_pointer_type (TREE_TYPE (enum_table_type)));
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (forset = build_decl (TYPE_DECL,
				   get_identifier ("__tmp_WIO_forset"),
				   result));
    DECL_SOURCE_LINE (forset) = 0;
    satisfy_decl (forset, 0);
#if 0
    /* build structure for powersets */
    decl1 = build_decl (FIELD_DECL, get_identifier ("data"),
			ptr_type_node);
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("bitlength"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("lowbound"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("name_table"),
			build_pointer_type (TREE_TYPE (enum_table_type)));
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("type"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_decl (FIELD_DECL, get_identifier ("__dummy2"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (forpowerset = build_decl (TYPE_DECL,
					get_identifier ("__tmp_IO_powerset"),
					result));
    DECL_SOURCE_LINE (forpowerset) = 0;
    satisfy_decl (forpowerset, 0);
#endif
    /* build the union */
    decl1 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forcharstring"),
			  TREE_TYPE (forcharstring)));
    listbase = decl1;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forset"),
			  TREE_TYPE (forset)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
#if 0
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forpowerset"),
			  TREE_TYPE (forpowerset)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
#endif
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forbyte"),
		  TREE_TYPE (lookup_name (
		    get_identifier ((ignore_case || ! special_UC) ? "byte" : "BYTE")))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forubyte"),
		  TREE_TYPE (lookup_name (
		    get_identifier ((ignore_case || ! special_UC)
				    ? "ubyte" : "UBYTE")))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forint"),
		  TREE_TYPE (lookup_name (
		    get_identifier ((ignore_case || ! special_UC)
				    ? "int" : "INT")))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__foruint"),
		  TREE_TYPE (lookup_name (
                    get_identifier ((ignore_case || ! special_UC) ?  "uint" : "UINT")))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forbool"),
		  TREE_TYPE (lookup_name (
		    get_identifier ((ignore_case || ! special_UC) ?  "bool" : "BOOL")))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forlong"), long_type));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forulong"),
		  ulong_type));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forchar"),
		  char_type_node));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forreal"),
		  TREE_TYPE (lookup_name (get_identifier
					  ((ignore_case || ! special_UC) ?  "real" : "REAL")))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forlongreal"),
		  TREE_TYPE (lookup_name (
                    get_identifier ((ignore_case || ! special_UC) ?  "long_real" : "LONG_REAL")))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forpointer"), ptr_type_node));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forinstance"),
		  TREE_TYPE (lookup_name (
		    get_identifier ((ignore_case || ! special_UC) ?  "instance" : "INSTANCE")))));
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    decl1 = grok_chill_variantdefs (NULL_TREE, listbase, NULL_TREE);
    TREE_CHAIN (decl1) = NULL_TREE;
    result = build_chill_struct_type (decl1);
    pushdecl (union_type = build_decl (TYPE_DECL,
			     get_identifier ("__tmp_WIO_union"),
				       result));
    DECL_SOURCE_LINE (union_type) = 0;
    satisfy_decl (union_type, 0);
    
    /* now build the final structure */
    decl1 = build_decl (FIELD_DECL, get_identifier ("__descr"),
			long_type);
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;

    decl2 = build_decl (FIELD_DECL, get_identifier ("__dummy1"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_decl (FIELD_DECL, get_identifier ("__t"),
			TREE_TYPE (union_type));
    
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (write_io_list_type = build_decl (TYPE_DECL,
					       get_identifier ("__tmp_WIO_list"),
					       result));
    DECL_SOURCE_LINE (write_io_list_type) = 0;
    satisfy_decl (write_io_list_type, 0);


    /* do the io-lits for READTEXT */
 build_read_io_list_type: ;

    if (read_io_list_type != NULL_TREE)
	/* already done */
      goto done;

    /* build type for charrange */
    decl1 = build_decl (FIELD_DECL, get_identifier ("ptr"),
			build_pointer_type (char_type_node));
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("lowbound"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("highbound"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (forcharrange = build_decl (TYPE_DECL,
					 get_identifier ("__tmp_IO_charrange"),
					 result));
    DECL_SOURCE_LINE (forcharrange) = 0;
    satisfy_decl (forcharrange, 0);
    
    /* build type for intrange */
    decl1 = build_decl (FIELD_DECL, get_identifier ("ptr"),
			ptr_type_node);
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("length"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("lowbound"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("highbound"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (forintrange = build_decl (TYPE_DECL,
					get_identifier ("__tmp_IO_intrange"),
					result));
    DECL_SOURCE_LINE (forintrange) = 0;
    satisfy_decl (forintrange, 0);
    
    /* build type for uintrange */
    decl1 = build_decl (FIELD_DECL, get_identifier ("ptr"),
			ptr_type_node);
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("length"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("lowbound"),
			ulong_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("highbound"),
			ulong_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (foruintrange = build_decl (TYPE_DECL,
					 get_identifier ("__tmp_IO_uintrange"),
					 result));
    DECL_SOURCE_LINE (foruintrange) = 0;
    satisfy_decl (foruintrange, 0);

    /* build type for set -- differs from that one is used for WRITETEXT */
    decl1 = build_decl (FIELD_DECL, get_identifier ("ptr"),
			ptr_type_node);
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("length"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("name_table"),
			build_pointer_type (TREE_TYPE (enum_table_type)));
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (forset = build_decl (TYPE_DECL,
				   get_identifier ("__tmp_RIO_forset"),
				   result));
    DECL_SOURCE_LINE (forset) = 0;
    satisfy_decl (forset, 0);
    
    /* build type for setrange */
    decl1 = build_decl (FIELD_DECL, get_identifier ("ptr"),
			ptr_type_node);
    DECL_INITIAL (decl1) = NULL_TREE;
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("length"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("name_table"),
			build_pointer_type (TREE_TYPE (enum_table_type)));
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("lowbound"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("highbound"),
			long_type);
    DECL_INITIAL (decl2) = NULL_TREE;
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (forsetrange = build_decl (TYPE_DECL,
					get_identifier ("__tmp_IO_forsetrange"),
					result));
    DECL_SOURCE_LINE (forsetrange) = 0;
    satisfy_decl (forsetrange, 0);

    /* build the union */
    decl1 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forcharstring"),
			  TREE_TYPE (forcharstring)));
    listbase = decl1;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forset"),
			  TREE_TYPE (forset)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
#if 0
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forpowerset"),
			  TREE_TYPE (forpowerset)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
#endif
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forcharrange"),
			  TREE_TYPE (forcharrange)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forintrange"),
			  TREE_TYPE (forintrange)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__foruintrange"),
			  TREE_TYPE (foruintrange)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forsetrange"),
			  TREE_TYPE (forsetrange)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forbyte"),
                  build_pointer_type (
		    TREE_TYPE (lookup_name (
		      get_identifier ((ignore_case || ! special_UC) ? "byte" : "BYTE"))))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forubyte"),
                  build_pointer_type (
		    TREE_TYPE (lookup_name (
		      get_identifier ((ignore_case || ! special_UC)
				      ? "ubyte" : "UBYTE"))))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forint"),
                  build_pointer_type (
		    TREE_TYPE (lookup_name (
		      get_identifier ((ignore_case || ! special_UC)
				      ? "int" : "INT"))))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__foruint"),
		  build_pointer_type (
		    TREE_TYPE (lookup_name (
                      get_identifier ((ignore_case || ! special_UC) ?  "uint" : "UINT"))))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forbool"),
                  build_pointer_type (
		    TREE_TYPE (lookup_name (
		      get_identifier ((ignore_case || ! special_UC) ?  "bool" : "BOOL"))))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;

    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forlong"), build_pointer_type (long_type)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forulong"),
		  build_pointer_type (ulong_type)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forchar"),
		  build_pointer_type (char_type_node)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forreal"),
                  build_pointer_type (
		    TREE_TYPE (lookup_name (
                      get_identifier ((ignore_case || ! special_UC) ?  "real" : "REAL"))))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forlongreal"),
                  build_pointer_type (
		    TREE_TYPE (lookup_name (
                      get_identifier ((ignore_case || ! special_UC) ?  "long_real" : "LONG_REAL"))))));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forpointer"), build_pointer_type (ptr_type_node)));
    TREE_CHAIN (decl1) = decl2;
    decl1 = decl2;
    
    decl2 = build_tree_list (NULL_TREE,
	      build_decl (FIELD_DECL,
		get_identifier ("__forinstance"),
                  build_pointer_type (
		    TREE_TYPE (lookup_name (
		      get_identifier ((ignore_case || ! special_UC) ?  "instance" : "INSTANCE"))))));
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    decl1 = grok_chill_variantdefs (NULL_TREE, listbase, NULL_TREE);
    TREE_CHAIN (decl1) = NULL_TREE;
    result = build_chill_struct_type (decl1);
    pushdecl (union_type = build_decl (TYPE_DECL,
			     get_identifier ("__tmp_RIO_union"),
				       result));
    DECL_SOURCE_LINE (union_type) = 0;
    satisfy_decl (union_type, 0);

    /* now build the final structure */
    decl1 = build_decl (FIELD_DECL, get_identifier ("__descr"),
			long_type);
    listbase = decl1;
    
    decl2 = build_decl (FIELD_DECL, get_identifier ("__t"),
			TREE_TYPE (union_type));
    TREE_CHAIN (decl1) = decl2;
    TREE_CHAIN (decl2) = NULL_TREE;
    
    result = build_chill_struct_type (listbase);
    pushdecl (read_io_list_type = build_decl (TYPE_DECL,
					      get_identifier ("__tmp_RIO_list"),
					      result));
    DECL_SOURCE_LINE (read_io_list_type) = 0;
    satisfy_decl (read_io_list_type, 0);

  done: ;
    maximum_field_alignment = save_maximum_field_alignment;
}

tree
build_chill_eoln (assoc)
    tree assoc;
{
  sorry ("");                  /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_existing (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_getassociation (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_gettextindex (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_gettextrecord (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_getusage (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_indexable (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_isassociated (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_outoffile (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_readable (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}


tree
build_chill_sequencible (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_variable (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

tree
build_chill_writeable (assoc)
    tree assoc;
{
  sorry ("");                   /* FIXME: write some code here */
  return error_mark_node;
}

static int
chill_stdio_p (text_arg)
    tree text_arg;
{
  tree tmp;
  int i;
  static char *predef_files [] =
    {
      "stdin",
      "stdout",
      "stderr",
      "STDIN",
      "STDOUT",
      "STDERR",
      0,
    };
  
  /* check if we have a LONG */
  if (TREE_TYPE (text_arg) != long_integer_type_node)
    return 0;
  
  if (TREE_CODE (text_arg) != VAR_DECL)
    return 0;
  
  /* check if we have a name */
  tmp = DECL_NAME (text_arg);
  if (tmp == NULL_TREE)
    return 0;
  
  /* check for name in predef_files */
  i = 0;
  while (predef_files[i])
    {
      if (!strcmp (IDENTIFIER_POINTER (tmp), predef_files[i])
	  && special_UC == isupper (predef_files[i][0]))
	return (1);
      i++;
    }
  return 0;
}


/* function walks through types as long as they are ranges,
   returns the type and min- and max-value form starting type.
*/

static tree get_final_type_and_range (item, low, high)
    tree  item;
    tree *low;
    tree *high;
{
    tree	wrk = item;
    
    *low = TYPE_MIN_VALUE (wrk);
    *high = TYPE_MAX_VALUE (wrk);
    while (TREE_CODE (wrk) == INTEGER_TYPE &&
	   TREE_TYPE (wrk) != NULL_TREE &&
	   TREE_CODE (TREE_TYPE (wrk)) == INTEGER_TYPE &&
	   TREE_TYPE (TREE_TYPE (wrk)) != NULL_TREE)
	wrk = TREE_TYPE (wrk);
    
    return (TREE_TYPE (wrk));
}


tree
build_chill_writetext (text_arg, format_str, exprlist)
     tree text_arg, format_str, exprlist;
{
  tree iolist_addr = null_pointer_node;
  tree iolist_length = integer_zero_node;
  tree fstr_addr;
  tree fstr_length;
  tree outstr_addr;
  tree outstr_length;
  tree fstrtype;
  tree outfunction;
  int  iolen;
  tree iolisttype;
  tree iolist;
  tree filename, linenumber;
  rtx  iolist_rtx = NULL_RTX;

  /* make some checks */
  if (text_arg == NULL_TREE || TREE_CODE (text_arg) == ERROR_MARK)
    return error_mark_node;
  if (format_str == NULL_TREE || TREE_CODE (format_str) == ERROR_MARK)
    return error_mark_node;
  
  if (exprlist != NULL_TREE)
    {
      if (TREE_CODE (exprlist) != TREE_LIST)
	return error_mark_node;
    }
  
  /* check the text argument */
  if (chill_varying_string_type_p (TREE_TYPE (text_arg)))
    {
      /* build outstr-addr and outstr-length assuming that this is a CHAR (n) VARYING */
      outstr_addr = force_addr_of (text_arg);
      outstr_length = size_in_bytes (CH_VARYING_ARRAY_TYPE (TREE_TYPE (text_arg)));
      outfunction = lookup_name (get_identifier ("_writestring"));
    }
  else if (chill_stdio_p (text_arg))
    {
      outstr_addr = text_arg;
      outstr_length = integer_zero_node;
      outfunction = lookup_name (get_identifier ("_writeunixstd"));
    }
  else
    {
      sorry ("WRITETEXT only implemented for CHARS (n) VARYING as output");
      sorry ("  or stdin, stdout or stderr");
      return error_mark_node;
    }
  
  /* check the format string */
  fstrtype = TREE_TYPE (format_str);
  if (CH_CHARS_TYPE_P (fstrtype) ||
      (flag_old_strings && TREE_CODE (format_str) == INTEGER_CST &&
       TREE_CODE (fstrtype) == CHAR_TYPE))
    {
      /* we have a character string */
      fstr_addr = force_addr_of (format_str);
      fstr_length = size_in_bytes (fstrtype);
    }
  else if (chill_varying_string_type_p (TREE_TYPE (format_str)))
    {
      /* we have a varying char string */
      fstr_addr
	= force_addr_of (build_component_ref (format_str, var_data_id));
      fstr_length = build_component_ref (format_str, var_length_id);
    }
  else
    {
      error ("`format string' for WRITETEXT must be a CHARACTER string");
      return error_mark_node;
    }
  
  /* process exprlist if present */
  if (exprlist != NULL_TREE)
    {
      tree	idxlist;
      int	idxcnt;
      
      iolen = list_length (exprlist);
      
      /* build indexlist for the io list */
      idxlist = build_tree_list (NULL_TREE,
		    build_chill_range_type (NULL_TREE,
					    integer_one_node,
					    build_int_2 (iolen, 0)));
      
      /* build the io-list type */
      iolisttype = build_chill_array_type (TREE_TYPE (write_io_list_type), 
					   idxlist, 0, -1);
      
      /* declare the iolist */
      iolist = build_decl (VAR_DECL, get_unique_identifier ("WRTEXT"), iolisttype);

      /* we want to get a variable which gets marked unused after
	 the function call, This is a little bit tricky cause the 
	 address of this variable will be taken and therefor the variable
	 gets moved out one level. However, we REALLY don't need this
	 variable again. Solution: push 2 levels and do pop and free
	 twice at the end. */
      push_temp_slots ();
      push_temp_slots ();
      iolist_rtx = assign_stack_temp (TYPE_MODE (TREE_TYPE (iolist)),
				      int_size_in_bytes (TREE_TYPE (iolist)), 0);
      DECL_RTL (iolist) = iolist_rtx;
      /* process the exprlist */
      idxcnt = 1;
      while (exprlist != NULL_TREE)
 	{
	  tree	item = TREE_VALUE (exprlist);
	  tree	idx = build_int_2 (idxcnt++, 0);
	  char	*fieldname = 0;
	  char	*enumname = 0;
	  tree	array_ref = build_chill_array_ref_1 (iolist, idx);
	  tree	item_type;
	  tree  range_low, range_high;
	  int	have_range = 0;
	  
	  /* next value in exprlist */
	  exprlist = TREE_CHAIN (exprlist);
	  if (item == NULL_TREE || TREE_CODE (item) == ERROR_MARK)
	    continue;

	  item_type = TREE_TYPE (item);
	  if (TREE_CODE (item_type) == REFERENCE_TYPE)
	    {
	      item_type = TREE_TYPE (item_type);
	      item = convert (item_type, item);
	    }

	  /* check for a range */
	  if (TREE_CODE (item_type) == INTEGER_TYPE &&
	      TREE_TYPE (item_type) != NULL_TREE)
	    {
	      /* we have a range. NOTE, however, on writetext we don't process ranges  */
	      item_type = get_final_type_and_range (item_type,
						    &range_low, &range_high);
	      have_range = 1;
	    }

	  if (TREE_CODE (item_type) == INTEGER_TYPE)
 	    {
	      int	type_size = TREE_INT_CST_LOW (TYPE_SIZE (item_type));
	      
	      switch (type_size)
  	        {
		case 8:
		  if (TREE_UNSIGNED (item_type))
		    {
		      enumname = "_IO_UByte";
		      fieldname = "__forubyte";
		    }
		  else
		    {
		      enumname = "_IO_Byte";
		      fieldname = "__forbyte";
		    }
		  break;
	      case 16:
		  if (TREE_UNSIGNED (item_type))
 		    {
		      enumname = "_IO_UInt";
		      fieldname = "__foruint";
		    }
		  else
		    {
		      enumname = "_IO_Int";
		      fieldname = "__forint";
		    }
		  break;
	      case 32:
	      try_long:
		  if (TREE_UNSIGNED (item_type))
		    {
		      enumname = "_IO_ULong";
		      fieldname = "__forulong";
		    }
		  else
		    {
		      enumname = "_IO_Long";
		      fieldname = "__forlong";
		    }
		  break;
	      default:
		  /* This kludge is because the lexer gives literals
		     the type long_long_{integer,unsigned}_type_node.  */
		  if (TREE_CODE (item) == INTEGER_CST)
		    {
		      if (int_fits_type_p (item, long_integer_type_node))
			{
			  item_type = long_integer_type_node;
			  item = convert (item_type, item);
			  goto try_long;
			}
		      if (int_fits_type_p (item, long_unsigned_type_node))
			{
			  item_type = long_unsigned_type_node;
			  item = convert (item_type, item);
			  goto try_long;
			}
		    }
  		  error ("Cannot process %d bits integer WRITETEXT argument %d.",
			 type_size, idxcnt + 1);
  		  continue;
	        }

	      if (fieldname)
 		{
		  tree	__t, __forxx;
		  
		  __t = build_chill_component_ref (array_ref,
						   get_identifier ("__t"));
		  __forxx = build_chill_component_ref (__t, get_identifier (fieldname));

		  expand_assignment (__forxx, item, 0, 0);
 		}
	      
 	    }
	  else if (TREE_CODE (item_type) == CHAR_TYPE)
	    {
		tree	__t, __forxx;

		__t = build_chill_component_ref (array_ref,
						 get_identifier ("__t"));
		__forxx = build_chill_component_ref (__t, get_identifier ("__forchar"));

		expand_assignment (__forxx, item, 0, 0);
		enumname = "_IO_Char";
	    }
	  else if (TREE_CODE (item_type) == BOOLEAN_TYPE)
	    {
	      tree	__t, __forxx;
	      
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier ("__forbool"));
	      expand_assignment (__forxx, item, 0, 0);
	      enumname = "_IO_Bool";
	    }
	  else if (TREE_CODE (item_type) == ENUMERAL_TYPE)
	    {
	      /* process an enum */
	      tree	table_name;
	      tree	__t, __forxx, __forxx_val, __forxx_ptr;
	      tree	context_of_type;
	      
	      /* determine the context of the type.
		 if TYPE_NAME (item_type) == NULL_TREE
		    if TREE_CODE (item) == INTEGER_CST
		       context = NULL_TREE -- this is wrong but should work for now
		    else
		       context = DECL_CONTEXT (item)
		 else
		    context = DECL_CONTEXT (TYPE_NAME (item_type)) */

	      if (TYPE_NAME (item_type) == NULL_TREE)
	      {
		  if (TREE_CODE (item) == INTEGER_CST)
		      context_of_type = NULL_TREE;
		  else
		      context_of_type = DECL_CONTEXT (item);
	      }
	      else
		  context_of_type = DECL_CONTEXT (TYPE_NAME (item_type));
	      
	      table_name = add_enum_to_list (item_type, context_of_type);
		  
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier ("__forset"));
	      __forxx_val = build_chill_component_ref (__forxx,
						       get_identifier ("value"));
	      __forxx_ptr = build_chill_component_ref (__forxx,
						       get_identifier ("name_table"));

	      expand_assignment (__forxx_val, item, 0, 0);
	      expand_assignment (__forxx_ptr, table_name, 0, 0);
	      enumname = "_IO_Set";
	    }
	  else if (TREE_CODE (item_type) == POINTER_TYPE)
	    {
	      /* we have a pointer */
	      tree	__t, __forxx;
	      
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier ("__forpointer"));
	      expand_assignment (__forxx, item, 0, 0);
	      enumname = "_IO_Pointer";
	    }
	  else if (item_type == instance_type_node)
	    {
	      /* we have an INSTANCE */
	      tree	__t, __forxx;
	      
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier ("__forinstance"));
	      expand_assignment (__forxx, item, 0, 0);
	      enumname = "_IO_Instance";
	    }
	  else if (TREE_CODE (item_type) == SET_TYPE && CH_BOOLS_TYPE_P (item_type))
	    {
	      tree	__t;
	      tree	the_location = item;

	      if (!CH_REFERABLE (item))
	        {
		  rtx   t;

		  /* we have to make a location to print that stuff */
		  the_location = build_decl (VAR_DECL,
					     get_unique_identifier ("WRTEXTBS"),
					     item_type);
		  t = assign_stack_temp (TYPE_MODE (item_type),
					 int_size_in_bytes (item_type), 0);
		  DECL_RTL (the_location) = t;
		  /* assign the value to it */
		  expand_assignment (the_location, item, 0, 0);
	        }
	      
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      if (CH_BOOLS_TYPE_P (item_type))
	        {
		  /* bitstring */
		  tree	__forxx, __forxx_string, __forxx_len;

		  __forxx = build_chill_component_ref (__t, get_identifier ("__forcharstring"));
		  __forxx_string = build_chill_component_ref (__forxx,
							      get_identifier ("string"));
		  __forxx_len = build_chill_component_ref (__forxx,
							   get_identifier ("string_length"));

		  expand_assignment (__forxx_string,
				     force_addr_of (the_location), 0, 0);
		  expand_assignment (__forxx_len,
				     build_chill_length (the_location), 0, 0);
		  enumname = "_IO_BitStr";
	        }
#if 0
	      else
	        {
		  /* must be a powerset */
		  tree	pslen = powersetlen (the_location);
		  tree	domain = TYPE_DOMAIN (item_type);
		  tree	lowbound = TYPE_MIN_VALUE (domain);
		  tree	__forxx, data, len, lb, nt, typ;
		  tree	table_ptr = null_pointer_node, context_of_type;
		  char	*type_name;
		  
		  __forxx = build_chill_component_ref (__t, get_identifier ("__forpowerset"));
		  data = build_chill_component_ref (__forxx,
						    get_identifier ("data"));
		  len = build_chill_component_ref (__forxx,
						   get_identifier ("bitlength"));
		  lb = build_chill_component_ref (__forxx,
						  get_identifier ("lowbound"));
		  nt = build_chill_component_ref (__forxx,
						  get_identifier ("name_table"));
		  typ = build_chill_component_ref (__forxx,
						   get_identifier ("type"));
		  
		  /* set up data, bitlength and lowbound */
		  expand_assignment (data, force_addr_of (the_location), 0, 0);
		  expand_assignment (len, pslen, 0, 0);
		  expand_assignment (lb, lowbound, 0, 0);
		  
                  if (TREE_CODE (domain) == INTEGER_TYPE &&
                      TREE_TYPE (domain) != NULL_TREE)
                    domain = get_final_type_and_range (domain, &range_low,
						       &range_high);
		  /* look for the type */
		  switch (TREE_CODE (domain))
		    {
		  case BOOLEAN_TYPE:
		      type_name = "_IO_Bool";
		      break;
		  case CHAR_TYPE:
		      type_name = "_IO_Char";
		      break;
		  case INTEGER_TYPE:
		      type_name = "_IO_Int";
		      break;
		  case ENUMERAL_TYPE:
		      type_name = "_IO_Set";
		      /* put enumeral in to list */
		      if (TYPE_NAME (domain) == NULL_TREE)
		        {
			  if (TYPE_NAME (item_type) == NULL_TREE)
			    {
			      context_of_type = DECL_CONTEXT (item);
			    }
			  else
			    {
			      context_of_type = DECL_CONTEXT (TYPE_NAME (item_type));
			    }
		        }
		      else
			  context_of_type = DECL_CONTEXT (TYPE_NAME (domain));
		      table_ptr = add_enum_to_list (domain, context_of_type);
		      break;
		  default:
		      error ("Cann't determine type of powerset.");
		      type_name = "_IO_UNUSED";
		      break;
		    }

		  /* set up name table and type of powerset */
		  expand_assignment (nt, table_ptr, 0, 0);
		  expand_assignment (typ,
				     lookup_name (get_identifier (type_name)), 0, 0);
		  enumname = "_IO_Powerset";
	        }
#endif
	    }
	  else if (chill_varying_string_type_p (item_type))
	    {
	      /* varying char string */
	      tree __t, __forxx, __forxx_string;
	      tree the_location = item;

	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier ("__forcharstring"));
	      __forxx_string = build_chill_component_ref (__forxx,
							  get_identifier ("string"));
	      if (!CH_REFERABLE (item))
		{
		  rtx t;
		  the_location = build_decl (VAR_DECL,
					     get_unique_identifier ("WRTEXTVS"),
					     item_type);
		  t = assign_stack_temp (TYPE_MODE (item_type),
					 int_size_in_bytes (item_type), 0);
		  DECL_RTL (the_location) = t;
		  expand_assignment (the_location, item,0 ,0);
		}
	      expand_assignment (__forxx_string,
				 force_addr_of (the_location), 0, 0);
	      enumname = "_IO_CharVarying";
	    }
	  else if (CH_CHARS_TYPE_P (item_type))
	    {
	      /* fixed character string */
	      tree __t, __forxx, __forxx_string, __forxx_len;
	      tree the_location = item;

	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier ("__forcharstring"));
	      __forxx_string = build_chill_component_ref (__forxx,
							  get_identifier ("string"));
	      __forxx_len = build_chill_component_ref (__forxx,
						       get_identifier ("string_length"));
	      if (!CH_REFERABLE (item))
		{
		  rtx   t;
		  the_location = build_decl (VAR_DECL,
					     get_unique_identifier ("WRTEXTCS"),
					     item_type);
		  t = assign_stack_temp (TYPE_MODE (item_type),
					 int_size_in_bytes (item_type), 0);
		  DECL_RTL (the_location) = t;
		  expand_assignment (the_location, item, 0, 0);
		}
	      expand_assignment (__forxx_string,
				 force_addr_of (the_location), 0, 0);
	      expand_assignment (__forxx_len,
				 size_in_bytes (item_type), 0, 0);
	      enumname = "_IO_CharStr";
	    }
	  else if (TREE_CODE (item_type) == REAL_TYPE)
	    {
	      /* process a (long_)real */
	      char	*field;
	      tree	__t, __forxx;
	      
	      if (lookup_name (ridpointers[RID_FLOAT]) == TYPE_NAME (item_type))
	        {
		  /* we have a real */
		  enumname = "_IO_Real";
		  field = "__forreal";
	        }
	      else
	        {
		  /* we have a long_real */
		  enumname = "_IO_LongReal";
		  field = "__forlongreal";
	        }
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier (field));
	      expand_assignment (__forxx, item, 0, 0);
	    }
	  else
 	    {
	      /* datatype is not yet implemented, issue a warning */
	      error ("cannot process mode of argument %d for WRITETEXT.", idxcnt + 1);
	      enumname = "_IO_UNUSED";
 	    }
	  
	  /* do assignment of the enum */
	  if (enumname)
 	    {
	      tree	descr = build_chill_component_ref (array_ref,
 					get_identifier ("__descr"));

	      expand_assignment (descr,
				 lookup_name (get_identifier (enumname)), 0, 0);
 	    }
 	}

      /* set up address and length of iolist */
      iolist_addr = build_chill_addr_expr (iolist, (char *)0);
      iolist_length = size_in_bytes (TREE_TYPE (iolist));
    }
  
  
    /* tree to call the function */

  filename = force_addr_of (get_chill_filename ());
  linenumber = get_chill_linenumber ();

  expand_expr_stmt (
      build_chill_function_call (outfunction,
	tree_cons (NULL_TREE, outstr_addr,
          tree_cons (NULL_TREE, outstr_length,
            tree_cons (NULL_TREE, fstr_addr,
              tree_cons (NULL_TREE, fstr_length,
                tree_cons (NULL_TREE, iolist_addr,
                  tree_cons (NULL_TREE, iolist_length,
		    tree_cons (NULL_TREE, filename,
		      tree_cons (NULL_TREE, linenumber,
                        NULL_TREE))))))))));

  /* get rid of the iolist variable, if we have one */
  if (iolist_rtx != NULL_RTX)
    {
      free_temp_slots ();
      pop_temp_slots ();
      free_temp_slots ();
      pop_temp_slots ();
    }

  /* return something the rest of the machinery can work with,
     i.e. (void)0 */
  return build1 (CONVERT_EXPR, void_type_node, integer_zero_node);
}

tree
build_chill_readtext (text_arg, format_str, exprlist)
     tree text_arg, format_str, exprlist;
{
  tree instr_addr, instr_length, infunction;
  tree fstr_addr, fstr_length, fstrtype;
  tree iolist_addr = null_pointer_node;
  tree iolist_length = integer_zero_node;
  tree filename, linenumber;
  rtx  iolist_rtx = NULL_RTX;
  
  /* make some checks */
  if (text_arg == NULL_TREE || TREE_CODE (text_arg) == ERROR_MARK)
    return error_mark_node;
  if (format_str == NULL_TREE || TREE_CODE (format_str) == ERROR_MARK)
    return error_mark_node;
  
  if (exprlist != NULL_TREE)
    {
      if (TREE_CODE (exprlist) != TREE_LIST)
	return error_mark_node;
    }
  
  /* check the text argument */
  if (CH_CHARS_TYPE_P (TREE_TYPE (text_arg)))
    {
      instr_addr = force_addr_of (text_arg);
      instr_length = size_in_bytes (TREE_TYPE (text_arg));
      infunction = lookup_name (get_identifier ("_readstring"));
    }
  else if (chill_varying_string_type_p (TREE_TYPE (text_arg)))
    {
      instr_addr
	= force_addr_of (build_component_ref (text_arg, var_data_id));
      instr_length = build_component_ref (text_arg, var_length_id);
      infunction = lookup_name (get_identifier ("_readstring"));
    }
  else if (chill_stdio_p (text_arg))
    {
      instr_addr = text_arg;
      instr_length = integer_zero_node;
      infunction = lookup_name (get_identifier ("_readunixstd"));
    }
  else
    {
      sorry ("READTEXT only implemented for CHARS(n), CHARS(n) VARYING");
      sorry ("  or stdin, stdout, stderr");
      return error_mark_node;
    }
  
  /* check the format string */
  fstrtype = TREE_TYPE (format_str);
  if (CH_CHARS_TYPE_P (fstrtype))
    {
      /* we have a character string */
      fstr_addr = force_addr_of (format_str);
      fstr_length = size_in_bytes (fstrtype);
    }
  else if (chill_varying_string_type_p (fstrtype))
    {
      /* we have a CHARS(n) VARYING */
      fstr_addr
	= force_addr_of (build_component_ref (format_str, var_data_id));
      fstr_length = build_component_ref (format_str, var_length_id);
    }
  else
    {
      error ("`format string' for READTEXT must be a CHARACTER string");
      return error_mark_node;
    }
  
  /* process exprlist if present */
  if (exprlist != NULL_TREE)
    {
      tree	idxlist, iolisttype, iolist;
      int	idxcnt, iolen;
      
      iolen = list_length (exprlist);
      
      /* build indexlist for the io list */
      idxlist = build_tree_list (NULL_TREE,
				 build_chill_range_type (NULL_TREE,
							 integer_one_node,
							 build_int_2 (iolen, 0)));
      
      /* build the io-list type */
      iolisttype = build_chill_array_type (TREE_TYPE (read_io_list_type),
					   idxlist, 0, -1);
      
      /* declare the iolist */
      iolist = build_decl (VAR_DECL, get_unique_identifier ("RDTEXT"), iolisttype);
      /* we want to get a variable which gets marked unused after
	 the function call */
      push_temp_slots ();
      push_temp_slots ();
      iolist_rtx = assign_stack_temp (TYPE_MODE (TREE_TYPE (iolist)),
				      int_size_in_bytes (TREE_TYPE (iolist)), 0);
      DECL_RTL (iolist) = iolist_rtx;
      
      /* process the exprlist */
      idxcnt = 1;
      while (exprlist != NULL_TREE)
	{
	  tree item = TREE_VALUE (exprlist);
	  tree idx = build_int_2 (idxcnt++, 0);
	  char *fieldname = 0;
	  char *enumname = 0;
	  tree array_ref = build_chill_array_ref_1 (iolist, idx);
	  tree item_type;
	  tree range_low = NULL_TREE, range_high = NULL_TREE;
	  int  have_range = 0;
	  tree item_addr = null_pointer_node;
	  
	  /* next value in exprlist */
	  exprlist = TREE_CHAIN (exprlist);
	  if (item == NULL_TREE || TREE_CODE (item) == ERROR_MARK)
	    continue;
	  
	  /* check for a range */
	  if (TREE_CODE (TREE_TYPE (item)) == INTEGER_TYPE &&
	      TREE_TYPE (TREE_TYPE (item)) != NULL_TREE)
	    {
	      /* we have a range */
	      item_type = get_final_type_and_range (TREE_TYPE (item),
						    &range_low, &range_high);
	      have_range = 1;
	    }
	  else
	    item_type = TREE_TYPE (item);
	  
	  if (TREE_READONLY (item_type))
	    error ("argument %d must not be READonly", idxcnt + 1);
	  if (!CH_REFERABLE (item))
	    error ("argument %d must be a location", idxcnt + 1);
	  else
	    /* build the pointer to item */
	    item_addr = force_addr_of (item);
	  
	  /* process the different types */
	  if (TREE_CODE (item_type) == INTEGER_TYPE)
	    {
	      int   type_size = TREE_INT_CST_LOW (TYPE_SIZE (item_type));
	      
	      if (have_range)
		{
		  /* do it for a range */
		  tree __t, __forxx, __ptr, __len, __low, __high;
		  
		  if (TREE_UNSIGNED (item_type))
		    {
		      enumname = "_IO_UIntRange";
		      fieldname = "__foruintrange";
		    }
		  else
		    {
		      enumname = "_IO_IntRange";
		      fieldname = "__forintrange";
		    }
		  
		  /* set up access to structure */
		  __t = build_chill_component_ref (array_ref,
						   get_identifier ("__t"));
		  __forxx = build_chill_component_ref (__t, get_identifier (fieldname));
		  __ptr = build_chill_component_ref (__forxx, get_identifier ("ptr"));
		  __len = build_chill_component_ref (__forxx, get_identifier ("length"));
		  __low = build_chill_component_ref (__forxx, get_identifier ("lowbound"));
		  __high = build_chill_component_ref (__forxx, get_identifier ("highbound"));
		  
		  /* do the assignments */
		  expand_assignment (__ptr, item_addr, 0, 0);
		  expand_assignment (__len, size_in_bytes (item_type), 0, 0);
		  expand_assignment (__low, range_low, 0, 0);
		  expand_assignment (__high, range_high, 0, 0);
		  fieldname = 0;
		}
	      else
		{
		  /* no range */
		  switch (type_size)
		    {
		    case 8:
		      if (TREE_UNSIGNED (item_type))
			{
			  enumname = "_IO_UByte";
			  fieldname = "__forubyte";
			}
		      else
			{
			  enumname = "_IO_Byte";
			  fieldname = "__forbyte";
			}
		      break;
		    case 16:
		      if (TREE_UNSIGNED (item_type))
			{
			  enumname = "_IO_UInt";
			  fieldname = "__foruint";
			}
		      else
			{
			  enumname = "_IO_Int";
			  fieldname = "__forint";
			}
		      break;
		    case 32:
		      if (TREE_UNSIGNED (item_type))
			{
			  enumname = "_IO_ULong";
			  fieldname = "__forulong";
			}
		      else
			{
			  enumname = "_IO_Long";
			  fieldname = "__forlong";
			}
		      break;
		    default:
		      error ("Cannot process %d bits integer for READTEXT argument %d.",
			     type_size, idxcnt + 1);
		      continue;
		    }
		}
	      
	      if (fieldname)
		{
		  tree __t, __forxx;
		  
		  __t = build_chill_component_ref (array_ref,
						   get_identifier ("__t"));
		  __forxx = build_chill_component_ref (__t, get_identifier (fieldname));
		  
		  expand_assignment (__forxx, item_addr, 0, 0);
		}
	    }
	  else if (TREE_CODE (item_type) == ENUMERAL_TYPE)
	    {
	      /* process an enum */
	      tree	table_name;
	      tree	__t, __forxx;
	      tree	context_of_type;
	      tree	__ptr, __len, __tab, __low = NULL_TREE, __high = NULL_TREE;
	      
	      /* determine the context of the type */
	      if (TYPE_NAME (item_type) == NULL_TREE)
		{
		  if (TREE_CODE (item) == INTEGER_CST)
		    context_of_type = NULL_TREE;
		  else
		    context_of_type = DECL_CONTEXT (item);
		}
	      else
		context_of_type = DECL_CONTEXT (TYPE_NAME (item_type));
	      
	      table_name = add_enum_to_list (item_type, context_of_type);
	      
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      if (have_range)
		{
		  __forxx = build_chill_component_ref (__t, get_identifier ("__forsetrange"));
		  __low = build_chill_component_ref (__forxx, get_identifier ("lowbound"));
		  __high = build_chill_component_ref (__forxx, get_identifier ("highbound"));
		  enumname = "_IO_SetRange";
		}
	      else
		{
		  __forxx = build_chill_component_ref (__t, get_identifier ("__forset"));
		  enumname = "_IO_Set";
		}
	      __ptr = build_chill_component_ref (__forxx, get_identifier ("ptr"));
	      __len = build_chill_component_ref (__forxx, get_identifier ("length"));
	      __tab = build_chill_component_ref (__forxx, get_identifier ("name_table"));
	      
	      /* do the assignments */
	      expand_assignment (__ptr, item_addr, 0, 0);
	      expand_assignment (__len, size_in_bytes (TREE_TYPE (item)), 0, 0);
	      expand_assignment (__tab, table_name, 0, 0);
	      if (__low != NULL_TREE)
		expand_assignment (__low, range_low, 0, 0);
	      if (__high != NULL_TREE)
		expand_assignment (__high, range_high, 0, 0);
	    }
	  else if (chill_varying_string_type_p (item_type))
	    {
	      /* varying character string */
	      tree	__t, __forxx, straddr, strlen;
	      
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier ("__forcharstring"));
	      straddr = build_chill_component_ref (__forxx, get_identifier ("string"));
	      strlen = build_chill_component_ref (__forxx, get_identifier ("string_length"));
	      
	      expand_assignment (straddr, item_addr, 0, 0);
	      expand_assignment (strlen,
		     size_in_bytes (TREE_TYPE (TREE_CHAIN (TYPE_FIELDS (item_type)))), 0 ,0);
	      enumname = "_IO_CharVarying";
	    }
	  else if (TREE_CODE (item_type) == SET_TYPE && CH_BOOLS_TYPE_P (item_type))
            {
	      /* we have a bitstring */
	      tree    __t, __forxx, straddr, strlen;
	      
	      __t = build_chill_component_ref (array_ref,
					       get_identifier ("__t"));
	      __forxx = build_chill_component_ref (__t, get_identifier ("__forcharstring"));
	      straddr = build_chill_component_ref (__forxx, get_identifier ("string"));
	      strlen = build_chill_component_ref (__forxx, get_identifier ("string_length"));
	      
	      expand_assignment (straddr, item_addr, 0, 0);
	      expand_assignment (strlen, build_chill_length (item), 0, 0);
	      enumname = "_IO_BitStr";
            }
	  else
	    {
	      /* datatype is not yet implemented, issue a warning */
	      warning ("cannot process argument %d for READTEXT yet.", idxcnt + 1);
	      enumname = "_IO_UNUSED";
	    }
	  
	  /* do the assignment of the enum */
	  if (enumname)
	    {
	      tree descr = build_chill_component_ref (array_ref,
						      get_identifier ("__descr"));
	      
	      expand_assignment (descr,
				 lookup_name (get_identifier (enumname)), 0, 0);
	    }
	}
      
      /* set up address and length of iolist */
      iolist_addr = build_chill_addr_expr (iolist, (char *)0);
      iolist_length = size_in_bytes (TREE_TYPE (iolist));
    }
  
  /* build the function call */
  filename = force_addr_of (get_chill_filename ());
  linenumber = get_chill_linenumber ();
  expand_expr_stmt (
     build_chill_function_call (infunction,
	tree_cons (NULL_TREE, instr_addr,
	   tree_cons (NULL_TREE, instr_length,
	      tree_cons (NULL_TREE, fstr_addr,
		 tree_cons (NULL_TREE, fstr_length,
		    tree_cons (NULL_TREE, iolist_addr,
		       tree_cons (NULL_TREE, iolist_length,
			  tree_cons (NULL_TREE, filename,
			     tree_cons (NULL_TREE, linenumber,
				NULL_TREE))))))))));
  
  /* get rid of the iolist variable, if we have one */
  if (iolist_rtx != NULL_RTX)
    {
      free_temp_slots ();
      pop_temp_slots ();
      free_temp_slots ();
      pop_temp_slots ();
    }
  
  /* return something the rest of the machinery can work with,
     i.e. (void)0 */
  return build1 (CONVERT_EXPR, void_type_node, integer_zero_node);
}

/* this function build all neccesary enum-tables used for
   WRITETEXT or READTEXT of an enum */

void build_enum_tables ()
{
    SAVE_ENUM_NAMES	*names;
    SAVE_ENUMS		*wrk;
    void		*saveptr;
    /* We temporarily reset the maximum_field_alignment to zero so the
       compiler's init data structures can be compatible with the
       run-time system, even when we're compiling with -fpack. */
    extern int maximum_field_alignment;
    int save_maximum_field_alignment;
    
    if (pass == 1)
	return;

    save_maximum_field_alignment = maximum_field_alignment;
    maximum_field_alignment = 0;

    /* output all names */
    names = used_enum_names;
    
    while (names != (SAVE_ENUM_NAMES *)0)
    {
	tree	var = get_unique_identifier ("ENUMNAME");
	tree	type;
	
	type = build_string_type (char_type_node,
				  build_int_2 (IDENTIFIER_LENGTH (names->name) + 1, 0));
	names->decl = decl_temp1 (var, type, 1,
				  build_chill_string (IDENTIFIER_LENGTH (names->name) + 1,
						      IDENTIFIER_POINTER (names->name)),
				  0, 0);
	names = names->forward;
    }

    /* output the tables and pointers to tables */
    wrk = used_enums;
    while (wrk != (SAVE_ENUMS *)0)
    {
	tree	varptr = wrk->ptrdecl;
	tree	table_addr = null_pointer_node;
	tree	init = NULL_TREE, one_entry;
	tree	table, idxlist, tabletype, addr;
	SAVE_ENUM_VALUES	*vals;
	int	i;
	
	vals = wrk->vals;
	for (i = 0; i < wrk->num_vals; i++)
	{
	    addr = build_chill_arrow_expr (
		      build_chill_array_ref (vals->name->decl,
                          tree_cons (NULL_TREE, integer_zero_node, NULL_TREE)), 0);
	    one_entry = tree_cons (NULL_TREE, build_int_2 (vals->val, 0),
				   tree_cons (NULL_TREE, addr, NULL_TREE));
	    one_entry = build_nt (CONSTRUCTOR, NULL_TREE, one_entry);
	    init = tree_cons (NULL_TREE, one_entry, init);
	    vals++;
	}

	/* add the terminator (name = null_pointer_node) to constructor */
	one_entry = tree_cons (NULL_TREE, integer_zero_node,
			       tree_cons (NULL_TREE, null_pointer_node, NULL_TREE));
	one_entry = build_nt (CONSTRUCTOR, NULL_TREE, one_entry);
	init = tree_cons (NULL_TREE, one_entry, init);
	init = nreverse (init);
	init = build_nt (CONSTRUCTOR, NULL_TREE, init);

	/* generate table */
	idxlist = build_tree_list (NULL_TREE,
		      build_chill_range_type (NULL_TREE,
					      integer_zero_node,
					      build_int_2 (wrk->num_vals, 0)));
	tabletype = build_chill_array_type (TREE_TYPE (enum_table_type),
					    idxlist, 0, -1);
	table = decl_temp1 (get_unique_identifier ("ENUMTAB"), tabletype,
			    1, init, 0, 0);
	table_addr = build_chill_arrow_expr (
			 build_chill_array_ref (table,
                             tree_cons (NULL_TREE, integer_zero_node, NULL_TREE)), 0);

	/* generate pointer to table */
	decl_temp1 (DECL_NAME (varptr), build_pointer_type (TREE_TYPE (enum_table_type)),
		    1, table_addr, 0, 0);

	/* free that stuff */
	saveptr = wrk->forward;
	
	free (wrk->vals);
	free (wrk);
	
	/* next enum */
	wrk = saveptr;
    }

    /* free all the names */
    names = used_enum_names;
    while (names != (SAVE_ENUM_NAMES *)0)
    {
	saveptr = names->forward;
	free (names);
	names = saveptr;
    }

    used_enums = (SAVE_ENUMS *)0;
    used_enum_names = (SAVE_ENUM_NAMES *)0;
    maximum_field_alignment = save_maximum_field_alignment;
}
