#! /bin/sh
#
# This script summarizes the regression-testing status of
# the CHILL compiler by running the regression tests and
# summarizing the results. It leaves its previous output
# in $LAST and the current output in $CURR.
#
# Note: This script is *NOT* for release to clients, because
# it acknowledges the existence of test cases which we cannot
# ship to them at this time.
#
# Command-line options:
#   -d, -D         # Turn on modest debugging msgs
#   -m, -M         # Suppress re-running the tests
#   -p, -P         # Print the PRMS correlation report
#   -u, -U         # Enable the cvs update/building of a fresh compiler
#   --srcdir=*     # Specify where the CHILL sources are
#   -O, -O2        # Request the CHILL compiler to optimize
# 
program=$0

# Default --srcdir to the directory where the script is found, 
# if a directory was specified.
# The second sed call is to convert `.//configure' to `./configure'.
srcdir=`echo $0 | sed 's|//|/|' | sed 's|/[^/]*$||'`
if [ x$srcdir = x$0 ]
then
  srcdir=.
fi

# If we're in the CHILL front-end directory, 
# move up to its devo level.
if [ -f ${srcdir}/ch-tree.h ]
then
  srcdir=`cd $srcdir/../..; pwd`
fi

objdir=`cd ../..; pwd`       # Where object files are
VERS=v3.4

# These flags are settable from the command line:
DEBUG=0         # 1=debug msgs enabled, rm's disabled
CVSUPDATE=0     # 1=do cvs update first
NOMAKES=0       # 1=suppress running tests, just report
PRMSRPT=1       # 1=print the PRMS correlation report
OPT=""          # CHILL compiler optimization flag

# Clear the totals
PRT=0           # total # of PRs
PRS=0           # # of PRs which PASSed
PRF=0           # # of PRs which FAILed
TTT=0           # total # of Tiny Test Cases
TTS=0           # # of TTCs which PASSed
TTF=0           # # of PRs which FAILed

set_paths ()
{
  MAKELOG=make.log${OPT}
  LAST=${objdir}/gcc/ch/.summary-1
  CURR=${objdir}/gcc/ch/.summary-0
  PRINTF=${objdir}/gcc/ch/utils/printf
  MAKE=`if [ -f /usr/latest/bin/make ] ; then echo /usr/latest/bin/make ;  else echo make ; fi`
  GAWK=`if [ -f /usr/unsupported/bin/gawk ] ; then echo /usr/unsupported/bin/gawk ;  else echo awk ; fi`
  echo srcdir=$srcdir objdir=$objdir
}

# Get local copy of the latest version, build it
cvs_update () 
{
  # Build the CHILL compiler
  cd ${srcdir}/gcc
  cvs update .

  cd ${objdir}
  ${srcdir}/configure --srcdir=${srcdir}
  cd ${objdir}/gcc
  ${MAKE} LANGUAGES='CHILL'

  # Build the CHILL runtime stuff
  cd ${objdir}/gcc/ch/runtime
  ${MAKE} libchill.a
}

onemake ()
{
  if [ $NOMAKES -eq 0  -o ! -f $MAKELOG ]
  then
    if [ $DEBUG -ne 0 ]
    then
      echo "Running tests in $1"
    fi
    cd $1
    if [ -s $MAKELOG ]
    then
      mv $MAKELOG $MAKELOG.old
    fi
    ${MAKE} clean
    ${MAKE} OPT=${OPT} -k check 1> $MAKELOG 2>&1
  fi
}


# Avoid divison by zero...
# $1 = passes, $2 = total number of tests
percentage ()
{
  if [ $2 -ne 0 ]
  then
    P=`expr '(' $1 '*' 100 ')' / $2`
  else
    P=0
  fi
}


dototals ()
{
  $PRINTF "\n" >> $CURR
  percentage $PRS $PRT
  $PRINTF "totals:\n"                >> $CURR
  $PRINTF "PR        %3d      %3d        %3d         %3d%%\n" $PRS $PRF $PRT $P >> $CURR

  percentage $TTS $TTT
  $PRINTF "TTC       %3d      %3d        %3d         %3d%%\n" $TTS $TTF $TTT $P >> $CURR
  $PRINTF "\ntotal     %3d      %3d        %3d         %3d%%\n" \
    `expr $TTS + $PRS` `expr $TTF + $PRF` `expr $TTT + $PRT` $P >> $CURR
  $PRINTF "                              compliance %% ^^^\n" >> $CURR
}


#############################################
# Fields on the output line:
# S=count of successful tests
# F=count of failed tests
# T=total # of tests
# P=percentage of success

printline ()
{
  if [ $T -eq 0 ]
  then
    T=`expr $S + $F`
  fi
  if [ $T -ne 0 ]
  then
    P=`expr '(' $S '*' 100 ')' / $T`
    if [ "$1" = "P" ]
    then
      $PRINTF "PR        %3d      %3d        %3d         %3d%%\n" $S $F $T $P >> $CURR
    elif [ "$1" = "T" ]
    then
      $PRINTF "TTC       %3d      %3d        %3d         %3d%%\n" $S $F $T $P >> $CURR
    else
      $PRINTF "doreport: parameter 1 incorrect: $1\n"
    fi
  fi
  if [ "$1" = "P" ]
  then
    PRT=`expr $PRT + $T`
    PRS=`expr $PRS + $S`
    PRF=`expr $PRF + $F`
  else
    TTT=`expr $TTT + $T`
    TTS=`expr $TTS + $S`
    TTF=`expr $TTF + $F`
  fi
}

#
# This function requires the $LOG and $DATA arguments
# It computes the totals.
#
dodirectory () 
{
  if [ -d ${objdir}/gcc/ch/testsuite/$1 ]
  then
    $PRINTF "$1:\n" >> $CURR

    cd ${objdir}/gcc/ch/testsuite/$1

    # Maybe clean the directory then run the tests
    onemake ${objdir}/gcc/ch/testsuite/$1

    grep '^[-A-Za-z0-9_][-A-Za-z0-9_]* PASS' $MAKELOG | \
      grep -v echo | sort -u > $$.PASS
    grep '^[-A-Za-z0-9_][-A-Za-z0-9_]* FAIL' $MAKELOG | \
      grep -v echo | sort -u > $$.FAIL

    S=`sed -n -e '/^pr-.... /p' -e '/^pr-...._ /p' \
       -e '/^pr-....-abort/p' < $$.PASS | wc -l` 
    F=`sed -n -e '/^pr-.... /p' -e '/^pr-...._ /p' \
       -e '/^pr-....-abort/p' < $$.FAIL | wc -l` 
    T=`/bin/ls -1 pr-????.dat pr-????_.dat pr-????-abort.dat \
       pr-????-gcov.dat 2> /dev/null | wc -l`
    printline "P" $S $F $T

    F=`sed -e '/^pr-.... /d' -e '/^pr-...._ /d' < $$.FAIL | wc -l`
    S=`sed -e '/^pr-.... /d' -e '/^pr-...._ /d' < $$.PASS | wc -l`
    X=`/bin/ls -1 *.dat 2> /dev/null | wc -l`
    V=`/bin/ls -1 pr-????.dat pr-????_.dat pr-????-abort.dat \
       pr-????-gcov.dat 2> /dev/null | wc -l`
    T=`expr $X - $V`
    printline "T" $S $F $T

    # Save pr stuff for checkprms function
    sed -n -e '/^pr-[0-9][0-9][0-9][0-9] /p' \
        -e '/^pr-[0-9][0-9][0-9][0-9]_ /p' < $$.PASS | \
      awk '{ print $1, $2; }' | \
      sed -e 's/pr-//'\
          -e 's/\([0-9][0-9][0-9][0-9]\)_ /\1/' >> ${objdir}/gcc/ch/prms.$$.PASS
    sed -n -e '/^pr-[0-9][0-9][0-9][0-9] /p' \
        -e '/^pr-[0-9][0-9][0-9][0-9]_ /p' < $$.FAIL | \
      awk '{ print $1, $2; }' | \
      sed -e 's/pr-//'\
          -e 's/\([0-9][0-9][0-9][0-9]\)_ /\1/' >> ${objdir}/gcc/ch/prms.$$.FAIL
    rm -f $$.PASS $$.FAIL
  fi
}

doreport () 
{
  cd ${objdir}/gcc/ch
  COMPILER=`grep '^echo host=' ../config.status`
  COMPILER=`echo $COMPILER | sed -e 's/^echo //' -e 's/ build=.*$//'`
  if [ -s $CURR ]
  then
    mv  $CURR $LAST
    $PRINTF "\n"
  fi

  $PRINTF "\nRegression Test Status:\n"                    >  $CURR 
  $PRINTF "              GNU CHILL Compiler\n"             >> $CURR
  $PRINTF "$COMPILER\n"                                    >> $CURR
  $PRINTF "`date`        report $VERS\n"                   >> $CURR
  $PRINTF "Type      Pass     Fail      Total       %% Pass\n\n" >> $CURR

  dodirectory execute
  dodirectory execute/oe
  dodirectory compile/elektra
  dodirectory compile/votrics
  dodirectory execute/telebras
  dodirectory compile
  dodirectory noncompile
  dodirectory examples

  dototals

  $PRINTF "\n" >> $CURR
}


checkprms () 
{
  cd ${objdir}/gcc/ch

  rm -f prms.$$.out

  # Get PR number and state, one per line
  query-pr --category=chill | \
  ${GAWK} -f ${srcdir}/gcc/ch/regression.awk1 | sort +0n > prms.$$.DB
  PRcount="`wc -l prms.$$.DB`"
  $PRINTF "These are the %d CHILL PRs:\n\n" $PRcount >> prms.$$.out
  ${srcdir}/gcc/ch/regression.prpt prms.$$.DB >> prms.$$.out

  sort +0n -u -o prms.$$.PASS prms.$$.PASS
  if [ -s prms.$$.PASS ]
  then
    $PRINTF "\n"              >> prms.$$.out
    $PRINTF "The following %d PRs report PASS:\n" `wc -l prms.$$.PASS` >> prms.$$.out
    $PRINTF "\n"              >> prms.$$.out
    ${srcdir}/gcc/ch/regression.prpt prms.$$.PASS >> prms.$$.out
  fi
  sort +0n -u -o prms.$$.FAIL prms.$$.FAIL
  if [ -s prms.$$.FAIL ]
  then
    $PRINTF "\n"              >> prms.$$.out
    $PRINTF "The following %d PRs report FAIL:\n" `wc -l prms.$$.FAIL` >> prms.$$.out
    $PRINTF "\n"              >> prms.$$.out
    ${srcdir}/gcc/ch/regression.prpt prms.$$.FAIL >> prms.$$.out
  fi

  if [ "$PRcount" -ne "0" ]
  then 
    ${GAWK} -f ${srcdir}/gcc/ch/regression.awk2 prms.$$.PASS prms.$$.FAIL prms.$$.DB > $$.out1
    if [ -s $$.out1 ]
    then
      $PRINTF "\n"        >> prms.$$.out
      $PRINTF "%d discrepancies between PRMS and the tests:\n" `wc -l $$.out1` >> prms.$$.out
      $PRINTF "\n"        >> prms.$$.out
      sort -u +1n $$.out1 >> prms.$$.out 
    fi
  fi
  rm -f prms.$$.PASS prms.$$.FAIL prms.$$.DB $$.out1
}

for arg in $*;
do
  case $next_arg in
  --srcdir)
    srcdir=$arg
    next_arg=
    ;;
  *)
    case $arg in
     -srcdir | --srcdir | --srcdi | --srcd | --src | --sr | --s)
	next_arg=--srcdir
	;;
     -srcdir=* | --srcdir=* | --srcdi=* | --srcd=* | --src=* | --sr=* | --s=*)
	srcdir=`echo $arg | sed 's/-*s[a-z]*=//'`
	;;

     -d | -D)         # Turn on modest debugging msgs
       DEBUG=1
       ;;
     -m | -M)         # Suppress re-running the tests
       NOMAKES=1      # Just generate the report from old data
       ;;
     -p | -P)         # Print the PRMS correlation report
       PRMSRPT=1
       ;;
     -u | -U)         # Enable the cvs update/building
       CVSUPDATE=1    # of a fresh compiler
       ;;
     -O | -O2)        # Request the CHILL compiler to optimize
       OPT=$arg
       ;;
     *)
       echo "usage: $program [-d][-m][-p][-u][-O][-O2] [--srcdir=path]" 
       exit 14
       ;;
     esac
  esac
done

cd ${objdir}/gcc/ch

set_paths

if [ $CVSUPDATE -eq 1 ]
then
  cvs_update
fi

# These are PR summary files for checkprms
rm -f ${objdir}/gcc/ch/prms.$$.PASS ${objdir}/gcc/ch/prms.$$.FAIL

doreport

cd ${objdir}/gcc/ch

# The historic info hasn't been too useful.
#cat $LAST $CURR >  outfile.$$
cat $CURR >  outfile.$$

if [ $PRMSRPT -eq 1 ]
then
  checkprms
  if [ -s prms.$$.out ]
  then
    cat prms.$$.out >> outfile.$$
  fi
  rm prms.$$.out
fi


# Output the explanatory notes
cat >> outfile.$$ << TpHaOiMnAeS

Notes:

  PR  = customer Problem Report
  TTC = Cygnus or contributed Tiny Test Case

TpHaOiMnAeS

mv outfile.$$ regression.latest
if [ $DEBUG -eq 0 ]
then
  mail bill < regression.latest
else
  cat regression.latest
fi

exit 0
