/* Implement Input/Output runtime actions for CHILL.
   Copyright (C) 1992,1993 Free Software Foundation, Inc.
   Author: Wilfried Moser, et al

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#define __CHILL_LIB__

#define DEBUG

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#ifdef sun	/* FIXME */
#include <floatingpoint.h>
#endif
#ifndef VMS
#include <unistd.h>
#endif

#include "iomodes.h"
#include "fileio.h"

/* define all needed exceptions */
EXCEPTION (associatefail);
EXCEPTION (spacefail);

/*
 * function move_file_name_okay
 *
*/

static int
move_file_name_okay (char *out, char *in, short in_len)
{
    if (in_len < MAXPATHLEN)
    {
	strncpy (out, in, in_len);
	out[in_len] = '\0';
	return (1);
    }
    else
	return (0);
} /* move_file_name_okay */

/*
 * function check_file_name
 *
*/

static int
check_file_name (char *name, __device *dev, int is_dir)
{
    int		i;
    
    /* check for blanks in name, remove trailing and leading blanks */
    while ((name[0] == ' ') || (name[0] == '\t'))
    {
	strcpy (name, name + 1);
	if (!name[0])
	    /* there are only blanks in string */
	    return (0);
    }
    
    i = strlen (name) - 1;
    while (i > 0)
    {
	if ((name[i] == ' ') || (name[i] == '\t'))
	    name[i] = '\0';
	else
	    break;
	i--;
    }
    if (i <= 0)
	return (0);
    if (strchr (name, ' ') || strchr (name, '\t') ||
	strchr (name, '*') || strchr (name, '%'))
	/* blanks or tabs in string */
	return (0);

    /* check for devices */
    if (!strcmp (name, "stdin"))
    {
	*dev = Stdin;
	return (1);
    }
    if (!strcmp (name, "stdout"))
    {
	*dev = Stdout;
	return (1);
    }
    if (!strcmp (name, "stderr"))
    {
	*dev = Stderr;
	return (1);
    }
    *dev = is_dir ? Directory : File;
    return (1);
} /* check_file_name */

/*
 * function check_name_in_table
 *
*/

static int
check_name_in_table (char *name, int shared_condition)
{
    int		index;
    
    for (index = 1; index < MAXUSEDFILES; index++)
    {
	if (__fs_table[index].ref_association)
	{
	    if (!strcmp (__fs_table[index].file_name, name) &&
		!shared_condition &&
		!__SHAREDBITSET (__fs_table[index].bit_list_ass))
		return (0);
	}
    }
    return (1);
} /* check_name_in_table */

/*
 * function associate_a_device
 *
*/

static void
associate_a_device (unsigned short attr, __device dev, int *handle,
				long *flen, short *nblocks)
{
    if (__DIRBITSET (attr))
	CAUSE (associatefail);
    
    *flen = 0;
    *nblocks = 0;
    switch (dev)
    {
    case Stdin:
	*handle = STDIN_FILENO;
	break;
    case Stdout:
	*handle = STDOUT_FILENO;
	break;
    case Stderr:
	*handle = STDERR_FILENO;
	break;
    default:
	CAUSE (associatefail);
	break;
    }
} /* associate_a_device */


/*
 * function associate_a_file
 *
*/

static void
associate_a_file (char *fname, int idx, short *attr,
			      int *fhdl, long *flen, short *nblocks)
{
    int		hdl;
    int		open_flags = 0;

    /* search for file already exists, otherwise clear existing bit */
    if ((hdl = open (fname, O_RDONLY)) < 0)
    {
	__CLREXISTINGBIT (*attr);
    }
    else
	close (hdl);

    if (__NEWBITSET (*attr))
    {
	/* new or overwrite */
	if (__NEWBITSET (*attr) &&
	    !(__EXISTINGBITSET (*attr)) &&
	    !(__SHAREDBITSET (*attr)))
	    /* overwrite or not existing and exclusive */
	    hdl = -1;
	else
	{
	    __fs_table[idx].status_word = EIO;
	    CAUSE (associatefail);
	}
    }
    else
    {
	/* old */
	if (__EXISTINGBITSET (*attr))
	{
	    /* calculate open flags */
	    if (__WRITEBITSET (*attr))
	    {
		if (__READBITSET (*attr))
		    /* now read/write */
		    open_flags = O_RDWR;
		else
		    /* write only */
		    open_flags = O_WRONLY;
	    }
	    else
	    {
		/* read only */
		open_flags = O_RDONLY;
	    }
	    
	    /* open the file */
	    if ((hdl = open (fname, open_flags)) < 0)
		CAUSE (associatefail);

	}
	else
	{
	    /* file doesn't exist -- error */
	    __fs_table[idx].status_word = ENOENT;
	    CAUSE (associatefail);
	}
    }
    
    /* return neccessary information */
    *fhdl = hdl;
    if (*nblocks > 16)
	*nblocks = 16;
    if (*nblocks < 1)
	*nblocks = 1;
    if (__EXISTINGBITSET (*attr))
    {
	/* get length of file */
	struct stat	sbuf;
	
	if (fstat (hdl, &sbuf) < 0)
	{
	    /* somethings wrong */
	    close (hdl);
	    CAUSE (associatefail);
	}
	*flen = sbuf.st_size;
    }
    else
	*flen = 0;
} /* associate_a_file */

/*
 * function _associate
 *
 * parameter:
 *  ass			pointer to __association location
 *  filename		name of `outside wold object'
 *  name_length		length of filename
 *  attributes		this bitlist is built by the compiler, basis for the
 *			bitlist are the given attributes in ASSOCIATE call.
 *			Further information is decribed in
 *			'CHILL - 286 FILE I/O IMPLEMENTATION DESCRIPTION'.
 *  number_of_blocks	defines the number of 512 byte blocks for the
 *			file buffer.
 *
 * returns:
 *  void
 *
 * exceptions:
 *  spacefail		file system table is full
 *  associatefail	bad attributes
 *
 * abstract:
 *  Associate an association location with an outside world object.
 *
*/

void
_associate (__association *ass, char *filename, short name_length,
	    unsigned short attributes, short number_of_blocks)
{
    int 	file_handle, index;
    long	length_of_file;
    char	name_of_file[MAXPATHLEN + 1];
    __device	device_desc;

    /* force ld to include debugging symbols of fstable.o ???
    FSDEBUG; */
    
    /* search for a free entry in the fs-table */
    /* start with one, 0 always reserved for file management */
    for (index = 1; index < MAXUSEDFILES; index++)
	if (__fs_table[index].ref_association == NULL)
	    break;
    if (index >= MAXUSEDFILES)
	CAUSE (spacefail);
    
    attributes &= 0xffe0;
    if (!__ISASSOCIATED (ass) &&
	(!(__DIRBITSET (attributes) && __FORMATBITSET (attributes))) &&
	(!(__SHAREDBITSET (attributes) && __WRITEBITSET (attributes))) &&
	(!__NEWBITSET (attributes) || __WRITEBITSET (attributes) ||
	   __DIRBITSET (attributes)) &&
	move_file_name_okay (name_of_file, filename, name_length) &&
	check_file_name (name_of_file, &device_desc, __DIRBITSET (attributes)) &&
	check_name_in_table (name_of_file, __SHAREDBITSET (attributes)))
    {
	/* set attribute existing */
	__SETEXISTINGBIT (attributes);
	
	switch (device_desc)
	{
	case File:
	    associate_a_file (name_of_file, index, &attributes, &file_handle,
			      &length_of_file, &number_of_blocks);
	    break;
	case Directory:
	    /* not implemented at the moment */
	    CAUSE (associatefail);
	    break;
	default:
	    associate_a_device (attributes, device_desc, &file_handle,
				&length_of_file, &number_of_blocks);
	    break;
	}

	ass->ref_access = NULL;
	ass->bitlist = attributes;
	ass->fileindex = index;
	__check_sum (ass);
	
	/* init fs-table entry */
	memset (&__fs_table[index], 0, sizeof (__fs_table_entry));
	__fs_table[index].ref_association = ass;
	__fs_table[index].bit_list_ass = attributes;
	__fs_table[index].file_name = __xmalloc_ (strlen (name_of_file) + 1);
	strcpy (__fs_table[index].file_name, name_of_file);
	__fs_table[index].device_script = device_desc;
	__fs_table[index].handle = file_handle;
	__fs_table[index].file_length = length_of_file;
	__fs_table[index].number_buffer_bytes = number_of_blocks * 512;
    }
    else
    {
	CAUSE (associatefail);
    }
    
} /* _associate */
