/* Implement Input/Output runtime actions for CHILL.
   Copyright (C) 1992,1993 Free Software Foundation, Inc.
   Author: Wilfried Moser, et al

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#define __CHILL_LIB__

#define DEBUG

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#ifdef sun	/* FIXME */
#include <floatingpoint.h>
#endif
#ifndef VMS
#include <unistd.h>
#endif

#include "iomodes.h"
#include "fileio.h"

/* define all needed exceptions */
EXCEPTION (empty);
EXCEPTION (protectionfail);
EXCEPTION (spacefail);

#ifndef VMS
#include <dirent.h>
#endif

#include <sys/stat.h>

#include "regexp.h"

#define PTRINC		100

#ifndef VMS
typedef int (*t_comp_func)(const void *, const void *);

/*
 * function sort_names
 *
*/

static int
sort_names (__ff_internal_file **arg1, __ff_internal_file **arg2)
{
    return (strcmp ((*arg1)->name, (*arg2)->name));
} /* sort_names */

#define BRKVAL	512
#define STORE(c)							\
    if ( ! (++cnt % BRKVAL))						\
	exp_ptr = __xrealloc_ (exp_ptr, BRKVAL * (cnt / BRKVAL + 1));	\
    exp_ptr[cnt - 1] = (c)

/*
 * function build_expr
 *
*/

static char *
build_expr (char *pattern)
{
    char	*exp_ptr;
    int		cnt = 0;
    
    exp_ptr = __xmalloc_ (BRKVAL);
    STORE ('^');
    while (1)
    {
	switch (*pattern)
	{
	case '\0':
	    STORE ('$');
	    STORE ('\0');
	    return (exp_ptr);
	    break;
	case '%':
	    STORE ('.');
	    break;
	case '*':
	    STORE ('.');
	    STORE ('*');
	    break;
	case '\\':
	case '[':
	case ']':
	case '^':
	case '$':
	case '(':
	case ')':
	case '.':
	case '|':
	case '+':
	case '?':
	    STORE ('\\');
	default:
	    STORE (*pattern);
	}
	pattern++;
    }
} /* build_expr */

#else
#include <descrip.h>
#include <rmsdef.h>
#endif

/*
 * function _find_file
 *
 * parameters:
 *  filename		filename with regular expression
 *  ffbuf		pointer to find_file_buffer
 *
 * returns:
 *  void
 *
 * exceptions:
 *  empty
 *  protectionfail
 *  spacefail
 *
 * abstract:
 *  initiate  a find-file operation and return the first filename
 *
*/

void
_find_file (TVaryingCharType *filename, __find_file_buffer *ffbuf)
{
#ifndef VMS
    extern char		*strrchr ();
    char		*wholename = alloca (filename->len + 1);
    char		*fname, *hdirname;
    TVaryingCharType	*dirname;
    TVaryingCharType	*curdir = alloca (sizeof (TVaryingCharType) + MAXPATHLEN);
    int			changedir = 0, dirnamelen;
    DIR			*dirp;
    struct dirent 	*dentry;
    struct stat		statb;
    __ff_internal_file	*wptr;
    regexp		*exp;
    
    if (!filename->len)
	CAUSE (empty);
    strncpy (wholename, &filename->body[0], filename->len);
    wholename[filename->len] = '\0';

    /* now figure out the directory part of the whole name */
    if ((fname = strrchr (wholename, '/')) == NULL)
    {
	/* no directory */
	fname = wholename;
	hdirname = "";
	dirnamelen = 0;
    }
    else
    {
	char	save_char = *++fname;
	
	*fname = '\0';
	dirname = alloca (sizeof (TVaryingCharType) + strlen (wholename) + 1);
	strcpy (&dirname->body[0], wholename);
	dirname->len = strlen (wholename);
	*fname = save_char;
	hdirname = &dirname->body[0];
	dirnamelen = strlen (hdirname);
	changedir = 1;
    }
    
    /* check for filename available */
    if (!strlen (fname))
	CAUSE (empty);

    /* compile the pattern */
    fname = build_expr (fname);
    _regerror_msg = NULL;
    exp = _regcomp (fname);
    free (fname);
    if (!exp)
	CAUSE (fileiofail);
    
    if (changedir)
    {
	/* get the current directory */
	_get_directory (curdir, MAXPATHLEN);
    
	/* now set to new directory -- exceptions will leave this function */
	_set_directory (dirname);
    }

    /* open directory */
    if ((dirp = opendir ("./")) == NULL)
    {
	/* can't open directory */
	if (changedir)
	    _set_directory (curdir);
	CAUSE (protectionfail);
    }

    /* set up internal structure */
    ffbuf->ptr = __xmalloc_ (sizeof (__ff_internal));
    memset (ffbuf->ptr, 0, sizeof (__ff_internal));
    
    /* get all files in the directory */
    while (dentry = readdir (dirp))
    {
	if (_regexec (exp, dentry->d_name, 1))
	{
	    /* we have a match */
	    if (stat (dentry->d_name, &statb) < 0)
	    {
		errno = 1;
		break;
	    }
	    
	    /* insert entry in table */
	    if (ffbuf->ptr->ptrs_allocated <= ffbuf->ptr->nr_files)
	    {
		if (!ffbuf->ptr->f)
		    ffbuf->ptr->f = __xmalloc_ (sizeof (char *) * PTRINC);
		else
		    ffbuf->ptr->f = __xrealloc_ (ffbuf->ptr->f,
						 (sizeof (char *) *
						  (ffbuf->ptr->ptrs_allocated +
						   PTRINC)));
		ffbuf->ptr->ptrs_allocated += PTRINC;
	    }
	    wptr = __xmalloc_ (sizeof (__ff_internal_file) + 
			       /* FIXME:  We use d_reclen rather than
				  d_namlen, because some systems don't 
				  have d_namlen.  Since d_reclen >= d_namlen
				  this is safe, though it wastes a few bytes
				  for each entry. */
			       dentry->d_reclen + dirnamelen + 1);
	    wptr->attrs = statb.st_mode;
	    wptr->time = statb.st_atime;
	    wptr->size = statb.st_size;
	    strcpy (wptr->name, hdirname);
	    strcat (wptr->name, dentry->d_name);
	    ffbuf->ptr->f[ffbuf->ptr->nr_files++] = wptr;
	    errno = 0;
	}
    }
    free (exp);
    if (errno)
    {
	/* we had an error -- return all allocated memory */
	int	i;
	
	for (i = 0; i < ffbuf->ptr->nr_files; i++)
	    free (ffbuf->ptr->f[i]);
	if (ffbuf->ptr->f)
	    free (ffbuf->ptr->f);
	free (ffbuf->ptr);
	ffbuf->ptr = NULL;
    }
    
    /* finished searching -- close directory and change directory back to original */
    closedir (dirp);
    if (changedir)
	_set_directory (curdir);

    /* if we had an error -- cause fileiofail */
    if (!ffbuf->ptr)
	CAUSE (fileiofail);
    
    /* if we had no file -- cause empty */
    if (!ffbuf->ptr->nr_files)
    {
	free (ffbuf->ptr);
	ffbuf->ptr = NULL;
	CAUSE (empty);
    }

    /* sort found entries */
    qsort (ffbuf->ptr->f, ffbuf->ptr->nr_files, sizeof (char *),
	   (t_comp_func)sort_names);
    
    /* return first file found */
    _find_next_file (ffbuf);
#else
    struct dsc$descriptor	fname = { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, NULL };
    struct dsc$descriptor	resname = { 0, DSC$K_DTYPE_T, DSC$K_CLASS_D, NULL };
    int				context = 0, status;
    struct stat			statb;
    
    fname.dsc$w_length = filename->len;
    fname.dsc$a_pointer = &filename->body[0];

    /* set up internal structure */
    ffbuf->ptr = __xmalloc_ (sizeof (__ff_internal));
    memset (ffbuf->ptr, 0, sizeof (__ff_internal));
    
    while ((status = lib$find_file (&fname, &resname, &context)) & 1)
    {
	char			*tmp;
	__ff_internal_file	*wptr;
	
	tmp = __xmalloc_ (resname.dsc$w_length + 1);
	strncpy (tmp, resname.dsc$a_pointer, resname.dsc$w_length);
	tmp[resname.dsc$w_length] = '\0';
	
	if (stat (tmp, &statb) < 0)
	{
	    free (tmp);
	    status = 1;
	    break;
	}

	/* insert entry in table */
	if (ffbuf->ptr->ptrs_allocated <= ffbuf->ptr->nr_files)
	{
	    if (!ffbuf->ptr->f)
		ffbuf->ptr->f = __xmalloc_ (sizeof (char *) * PTRINC);
	    else
		ffbuf->ptr->f = __xrealloc_ (ffbuf->ptr->f,
					     (sizeof (char *) *
					      (ffbuf->ptr->ptrs_allocated +
					       PTRINC)));
	    ffbuf->ptr->ptrs_allocated += PTRINC;
	}
	wptr = __xmalloc_ (sizeof (__ff_internal_file) + 
			   strlen (tmp) + 1);
	wptr->attrs = statb.st_mode;
	wptr->time = statb.st_atime;
	wptr->size = statb.st_size;
	strcpy (wptr->name, tmp);
	ffbuf->ptr->f[ffbuf->ptr->nr_files++] = wptr;
	free (tmp);
    }
    lib$find_file_end (&context);
    str$free1_dx (&resname);
    if (status != RMS$_NMF)
    {
	int		i;
	
	for (i = 0; i < ffbuf->ptr->nr_files; i++)
	    free (ffbuf->ptr->f[i]);
	if (ffbuf->ptr->f)
	    free (ffbuf->ptr->f);
	free (ffbuf->ptr);
	ffbuf->ptr = NULL;
    }

    /* if we had an error -- cause fileiofail */
    if (!ffbuf->ptr)
	CAUSE (fileiofail);
    
    /* if we had no file -- cause empty */
    if (!ffbuf->ptr->nr_files)
    {
	free (ffbuf->ptr);
	ffbuf->ptr = NULL;
	CAUSE (empty);
    }

    /* return first file found */
    _find_next_file (ffbuf);

#endif
} /* _find_file */
