/* Implement Input/Output runtime actions for CHILL.
   Copyright (C) 1992,1993 Free Software Foundation, Inc.
   Author: Wilfried Moser, et al

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#define __CHILL_LIB__

#define DEBUG

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#ifdef sun	/* FIXME */
#include <floatingpoint.h>
#endif
#ifndef VMS
#include <unistd.h>
#endif

#include "iomodes.h"
#include "fileio.h"

/* define all needed exceptions */
EXCEPTION (spacefail);
EXCEPTION (textfail);

/*
 * function writetext_io
 *
 * parameters:
 *  txt			pointer to text location
 *  idx			index (if indexed, else 0)
 *  formfeed		indicates presence of 'require formfeed symbol'
 *  act_max_index	number of bytes to write
 *  iotype		indicates the type of record boundary for the
 *			actual textv record
 *
 * returns:
 *  __format_result
 *
 * abstract:
 *  output function for _writetext
 *
*/

static __format_result
writetext_io (__text *txt, int idx, int *formfeed,
	      int *act_max_index, __io_type iotype)
{
    __fs_table_entry	*fse = &__fs_table[txt->fileindex];
    
    switch (iotype)
    {
    case io_1:
	__ACCSETIOT1 (fse->bit_list_acc);
	break;
    case io_2:
	__ACCSETIOT2 (fse->bit_list_acc);
	break;
    case io_3:
	__ACCSETIOT3 (fse->bit_list_acc);
	break;
    case io_4:
	__ACCSETIOT4 (fse->bit_list_acc);
	break;
    default:
	__ACCSETIOT5 (fse->bit_list_acc);
	break;
    }
    
    if (*formfeed)
	__ACCSETFF (fse->bit_list_acc);
    
    if (__ACCDYNAMIC (fse->bit_list_acc))
	/* dynamic records on file */
	*(long *)(txt->text_ptr) = *act_max_index;
    else
    {
	/* fill record with blanks */
	char	*tmp = txt->text_ptr + 4;
	int	i;
	
	for (i = *act_max_index; i < fse->rec_len_acc; i++)
	    tmp[i] = ' ';
	*(long *)(txt->text_ptr) = fse->rec_len_acc;
    }

#if 1
    abort();
#else
    /* This code should be re-written in Chill! FIXME! */
    /* call writerecord */
    TRY ((6, EXNAME(protectionfail), EXNAME(notconnected), EXNAME(writefail),
	  EXNAME(spacefail), EXNAME(rangefail), EXNAME(readfail)))
    {
	_writerecord (fse->ref_access, idx, txt->text_ptr);
    }
    ON (protectionfail)
	fse->index_diff = 1;
    ON (notconnected)
	fse->index_diff = 2;
    ON (spacefail)
	fse->index_diff = 3;
    ON (rangefail)
	fse->index_diff = 4;
    ON (readfail)
	fse->index_diff = 5;
    ON (writefail)
	fse->index_diff = 6;
    ENDON;
#endif
    
    *formfeed = 0;
    __ACCCLRIOT (fse->bit_list_acc);
    __check_sum (txt);
    return (fse->index_diff ? not_implemented : end_of_formatting);
} /* writetext_io */

/*
 * function _writetext
 *
 * parameters:
 *  txt			pointer to text location
 *  idx			index, if specified
 *  formatstr		pointer to format string
 *  formatstrlen	length of format string
 *  iolist		pointer to io-list
 *  iolistlen		length of io-list in bytes
 *
 * returns:
 *  void
 *
 * exceptions:
 *  spacefail		output string to short
 *  textfail		any error during formatting
 *
 * abstract:
 *  perform same function as `(f)printf' in C for CHILL
 *
*/

void
_writetext (__text *txt, int idx, char *formatstr,
	    int formatstrlen, __tmp_WIO_list *iolist,
	    int iolistlen)
{
    __fs_table_entry	*fse = &__fs_table[txt->fileindex];
    int			formfeed = 0;
    __format_result	res;
    int			iol_index = 0;
    
    if (!__ISCONNECTED (txt))
	CAUSE (notconnected);

    if (!__ACCWRITE (txt->bitlist))
	CAUSE (writefail);

    fse->index_diff = 0;

    /* call formatting function */
    res = __write_format (formatstr, formatstrlen,
			  iolist, iolistlen / sizeof (__tmp_WIO_list), &iol_index,
			  txt->text_ptr + 4, fse->rec_len_acc,
			  &fse->actual_index, &fse->max_act_index,
			  1, &formfeed, txt, idx, &writetext_io);
    txt->actual_index = fse->actual_index;
    *(long *)(txt->text_ptr) = fse->max_act_index;
    if (formfeed)
	__ACCSETFF (fse->bit_list_acc);
    else
	__ACCCLRFF (fse->bit_list_acc);
    if (res != end_of_formatting)
	CAUSE (textfail);
    switch (fse->index_diff)
    {
    case 0:
	break;
    case 1:
	CAUSE (protectionfail);
	break;
    case 2:
	CAUSE (notconnected);
	break;
    case 3:
	CAUSE (spacefail);
	break;
    case 4:
	CAUSE (rangefail);
	break;
    case 5:
	CAUSE (readfail);
	break;
    case 6:
	CAUSE (writefail);
	break;
    default:
	fprintf (stderr, "_writetext: index_diff out of range (%d)\n", fse->index_diff);
	abort ();
    }
} /* _writetext */
